// bbdmux12.cpp
//
// MPEG2AVI
// --------
//// v0.16B34
//   modified bs_getbytes() to allow fast skipping of bytes
//   if the parameter (*pcbStored ==NULL), skip N bytes
//   process_stream_payload() and process_substream_payload() take advantage
//   of this optimization
//  v0.16B33 upgraded to BBDMUX v1.3, v1.3 fixes AC3 stream demuxing, when
//   the AC3's substreamID != 0x80.  process_substream() contains the changes
//  v0.16B31 fixed reporting error in TextLogMsg() when log is purged
//       added log-size check for TextLogErr()
//  v0.16B3 (modified bbdbits.cpp, too), added logging via
//   TextLogErr() and TextLogMsg().
//   go_demux() now checks for EOF condition [bs_end_bs()], prior to running
//   through loop.
//  v0.16B23 added lba_seek_next_sysh( DWORD dwStartLBA );
//   Scans for SYSTEM_HEADER starting @ dwStartLBA.
//   The outside world should use this function instead of fseek_lba()
//  v0.16B22 added bs_getbytes() function.  The process_stream_payload()
//   and process_substream_payload() now utilize bs_getbytes() to
//   retrieve long runs of data.  If the data is byte-aligned, the retrieval
//   process is much faster.
//  v0.16B21 removed all exit() commands, to reduce eof errors
//  v0.16B2 complete re-write from BBDMUX v1.2, 
//    converted to C++ object
//  v0.14  minor cosmetic adjustments to stats reporting
//  v0.13  initial release
//
//  The original BBDMUX executed in a while() loop until EOF.
//  Now, BBDMUX has been modified into demux_mpeg2(), which executes
//  a single time (or until *some* data has been extracted) and then
//  exits.
//
//  demux_mpeg2() calls s_fwrite(), which is rerouted to the g_demuxbuffer
//  queue.  BBDMUX calls s_fwrite() many times (transfering
//  just 1 byte per call.)  Multiple calls to s_fwrite() can cause the
//  program to malfunction.
//
//  This is due to the queue's FILLTHRESHOLD policy.  During a single
//  iteration of demux_mpeg2(), the queue may become full.  Some 
//  s_fwrite()'s will succeed and some will fail.  Thus not all of the
//  stream data is properly transfered to the queue.  This is a
//  *catastrophic* malfunction.
//
//  See demuxbuf.cpp for the current workaround.
//  
//  MAJOR CHANGES SINCE v0.16B2 (c++ conversion)
// --------------------------------------------------
//
// starting with v0.16B2, the C++ version differs from the C version in
// several ways.
//   1) the primary demuxing routine has been broken into several
//      functions  go_demux(), process_substream(), process_stream_data()
//   2) the main function (go_demux) no longer decides which stream to
//      capture...it now passes ALL streams/substreams to the subfunctions.
//   3) The subfunctions then "filter" the stream_to_get and substream_to_get
//      (with a little more mods, bbdmux could rip *all* payload streams 
//       in a single pass.)
//   4) All of bbdmux's bitstream-manipulation functions are now prefixed
//      with "bs_...()"
//   5) The demuxer's state-variables are stored in struct demuxstate_struct m_s
//
//   5) go_demux( *dest_buffer, *dwBytesReturned) stores the demuxed-data
//      in dest_buffer, and records the #bytes demuxed in *dwBytesReturned
//      (originally, demux_mpeg2() would call s_fwrite(), which would in
//       in term call demux_buffer::push_data() )
//
//  There are three files for class cDemuxer : 
//   bbdmux12.cpp, bbdmux12.h, bbdbits.cpp


/*************************************************************************
	bbDMUX by Brent Beyeler, beyeler@home.com
*************************************************************************/

#include "bbdmux12.h"
#include <string.h>

#undef SHOW_RESYNCHS

#define MPEG_PROGRAM_END_CODE	 0x000001B9
#define PACK_START_CODE 		 0x000001BA
#define SYSTEM_HEADER_START_CODE 0x000001BB
#define PACKET_START_CODE_PREFIX 0x000001

#define PROGRAM_STREAM_MAP		 0xBC
#define PRIVATE_STREAM_1		 0xBD
#define PADDING_STREAM			 0xBE
#define PRIVATE_STREAM_2		 0xBF
#define ECM_STREAM				 0xF0
#define EMM_STREAM				 0xF1
#define PROGRAM_STREAM_DIRECTORY 0xFF
#define DSMCC_STREAM			 0xF2
#define ITUTRECH222TYPEE_STREAM  0xF8

#define VOB_AC3_SSLOW			 0x80
#define VOB_AC3_SSHIGH			 0x8F

#define TEMPSTR_LENGTH 1024



// demux state structure
// encapsulates all the variables that comprise the demultiplexor's state

// constructor
cDemuxer::cDemuxer( void )
{
	// clear these vars separately
	m_outfp = NULL;
	m_bs.fhandle = _NOFILE; // _FERROR = file error
	m_bs.qwLength = -1;	// set bitstream length to -1, no file open
	stream_to_get = STREAM_NOT_SET;
	substream_to_get = STREAM_NOT_SET;

	// housekeeping variables (text logs and error logs)
	// allocate memory for logs
	m_szErrorLog = new char[ LOG_LENGTH ];
	m_szMessageLog = new char[ LOG_LENGTH ];
	m_szText = new char[ TEXTMSG_LENGTH ];
	m_szText2 = new char[ TEXTMSG_LENGTH ];
	m_szLastLogErr = new char[ TEXTMSG_LENGTH ];

	m_lastError = 0;
	m_logCounter = 0;

	if ( (m_szErrorLog == NULL) || ( m_szMessageLog == NULL ) || 
		( m_szText == NULL ) || ( m_szText2 == NULL ) ||
		( m_szLastLogErr == NULL ) )
	{
		// catastrophic error, couldn't allocate logs
		fprintf( stderr, "\ncDemuxer::cDemuxer() : couldn't allocate memory.");
		// exit( ERROR );
	}

	// clear the logs
	m_szLastLogErr[ 0 ] = 0;  // NULL char
	m_szText2[ 0 ] = 0;  // NULL char
	m_szText[ 0 ] = 0;  // NULL char
	m_szMessageLog[ 0 ] = 0;  // NULL char
	m_szErrorLog[ 0 ] = 0;  // NULL char

	reset_demux();  // must call after we set m_bs.fhandle = _NOFILE

	TextLogMsg( LOG_NEW_ENTRY, "cDemuxer() : object construction done." );
}


// destructor
cDemuxer::~cDemuxer()
{
	if ( m_outfp != NULL )
		fclose( m_outfp );

	if ( BSOPEN( (&m_bs) ) )
	{
		print_summary();
		bs_close_bit_stream_r(&m_bs);
	}

	if ( m_szErrorLog != NULL )
		delete( m_szErrorLog );

	if ( m_szMessageLog != NULL )
		delete( m_szMessageLog );

	if ( m_szText != NULL )
		delete( m_szText );

	if ( m_szText2 != NULL )
		delete( m_szText2 );
}

// process_payload( stream_id, cbData )
// cbData = requested # bytes to read, from <stream_id>
// returns the actual# of bytes read ( less than or equal to cbData)
// updates the stream_bytes[ stream_id ] counter, but not stream_packets[]
unsigned long
cDemuxer::process_stream_payload( unsigned char stream_id, unsigned long cbData )
{
	unsigned long j = 0, dwrc = 0;	// dword return code
	unsigned long cbPayload = 0; // # desired payload bytes processed
	DWORD cbDataGrabbed = 0;	// return-value from bs_getbytes()

	if ( stream_to_get < 0 )
	{
		TextLogErr( DMXERR_FAIL, "process_stream_payload() : stream_to_get not initialized!" );
		return 0;
	}

	/*if (stream_id == stream_to_get)
	{	// this is the stream we want to store
		
		for (j = 0; j < cbData; j++)
		{
			byte = bs_getbits(&m_bs, 8);
			if (s_fwrite(&byte, 1, 1, m_outfp) != 1)
			{
				fprintf( stderr, "Error writing to output file %s.", m_bs.filename );
				return DMXERR_ENDOFSTREAM;
				//exit (1);
			}
		}
		cbPayload = j;
		// load bytes into output buffer
	}
	else
	{	// no some other stream, just ignore it
		for (j = 0; j < cbData; j++)
		{
			byte = bs_getbits(&m_bs, 8);
		}
		cbPayload = 0;
	}*/

//	dwrc = bs_getbytes( &m_bs, m_pBuffer, cbData, &cbDataGrabbed ); 
	if ( stream_id == stream_to_get )
	{
		// grab cbData bytes into -> m_pBuffer[]
		dwrc = bs_getbytes( &m_bs, m_pBuffer, cbData, &cbDataGrabbed ); 
		if ( dwrc == BSERR_OK )
		{
			// write the grabbed data to Demuxer's output buffer
			j = s_fwrite( m_pBuffer, 1, cbDataGrabbed, m_outfp);
			if ( j != cbDataGrabbed )
			{
				// uh oh, didn't write all bytes to output buffer
				sprintf( m_szText, "process_stream_payload() : couldn't write data(%ld) to cDemuxer buffer!",
					cbDataGrabbed );
				TextLogErr( DMXERR_FAIL, m_szText );
			}
			cbPayload = j; // return this# of bytes actually written into buffer
		}
		else
			cbPayload = 0;
	}	
	else
	{	// (stream_id != stream_to_get)
		// skip over the bytes
		dwrc = bs_getbytes( &m_bs, m_pBuffer, cbData, NULL ); 
		cbPayload = 0;
	}

	//s.stream_bytes[stream_id] += cbDataGrabbed; // update counter
	s.stream_bytes[stream_id] += cbData; // update counter

	return cbPayload;
}  // cDemuxer::process_payload( )



// process_substream_payload( stream_id, substream_id, cbData )
// cbData = requested # bytes to read, from <stream_id>
// returns the actual# of bytes read ( less than or equal to cbData)
// updates the stream_bytes[ stream_id ] counter, but not stream_packets[]
unsigned long
cDemuxer::process_substream_payload( unsigned char stream_id, 
	unsigned char substream_id, unsigned long cbData )
{
	unsigned long j = 0;
	unsigned long cbPayload = 0; // # desired payload bytes processed
	DWORD dwrc = 0;	// dword return code
	DWORD cbDataGrabbed = 0;	// return-value from bs_getbytes()

	if ( stream_to_get < 0 )
	{
		TextLogErr( DMXERR_FAIL, "process_substream_payload() : stream_to_get not initialized!" );
		return 0;
	}


	/*if ( stream_id == stream_to_get && substream_id == substream_to_get )
	{	// this is the stream we want to store
		for (j = 0; j < cbData; j++)
		{
			byte = bs_getbits(&m_bs, 8);
			if (s_fwrite(&byte, 1, 1, m_outfp) != 1)
			{
				fprintf( stderr, "Error writing to output file %s.", m_bs.filename );
				return DMXERR_ENDOFSTREAM;
				//exit (1);
			}
		}
		cbPayload = j;
		j = bs_getbytes( &m_bs, cbData ); // load bytes into output buffer
		cbPayload = j;
	}
	else
	{	// no some other stream, just ignore it
		for (j = 0; j < cbData; j++)
		{
			byte = bs_getbits(&m_bs, 8);
		}
		cbPayload = 0;
	}*/

	//dwrc = bs_getbytes( &m_bs, m_pBuffer, cbData, &cbDataGrabbed ); 
	if ( stream_id == stream_to_get && substream_id == substream_to_get )
	{
		// grab cbData bytes into -> m_pBuffer[]
		dwrc = bs_getbytes( &m_bs, m_pBuffer, cbData, &cbDataGrabbed ); 
		if ( dwrc == BSERR_OK )
		{
			// write the grabbed data to Demuxer's output buffer
			j = s_fwrite( m_pBuffer, 1, cbDataGrabbed, m_outfp);
			if ( j != cbDataGrabbed )
			{
				// uh oh, didn't write all bytes to output buffer
				sprintf( m_szText,"process_substream_payload() : couldn't write data(%ld) to cDemuxer buffer!",
					cbDataGrabbed );
				TextLogErr( DMXERR_FAIL, m_szText );
			}
			cbPayload = j; // return this# of bytes actually written into buffer
		}
		else
			cbPayload = 0;	// BSERR_FAIL
	}	
	else
	{	// don't want this stream, so discard the data
		dwrc = bs_getbytes( &m_bs, m_pBuffer, cbData, NULL ); 
		cbPayload = 0;
	}

	//s.stream_bytes[stream_id] += cbDataGrabbed;
	//s.substream_bytes[substream_id] += cbDataGrabbed;
	s.stream_bytes[stream_id] += cbData;
	s.substream_bytes[substream_id] += cbData;

	return cbPayload;
}	// cDemuxer::process_substream_payload( )


unsigned long
cDemuxer::process_substream( unsigned char  stream_id, unsigned char substream_id )
{
	unsigned long j;
	unsigned long bytes_to_get;
	unsigned long cbPayload = 0; // # payload bytes processed so far
	BOOL fAC3data = FALSE;

	if ( stream_to_get < 0 )
	{
		TextLogErr( DMXERR_FAIL, "process_substream() : m_s not initialized!" );
		return 0;
	}

//if (stream_id == stream_to_get)
//{
//	if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag) && 
//		substream_to_get != STREAM_NOT_SET )
//	{ // yes, this is a substream

	++ (s.substream_packets[substream_id] );

//	  if (substream_id == substream_to_get)
//	  {

	// v0.16B33, fixed AC3 substream decoding, the parser now recognizes
	// ac3 substreams of ID 0x80 through 0x8F
	if ( substream_id >= VOB_AC3_SSLOW && substream_id  <= VOB_AC3_SSHIGH )
	{
		fAC3data = TRUE; // set flag
		bs_getbits(&m_bs, 24);	// discard the 4 AC3 ident bytes
		s.stream_bytes[stream_id] += 3;
		s.substream_bytes[ substream_id ] += 3;
	}
	else
	{
		if ( substream_id == substream_to_get )
		{
			// manually write this byte to the output-buffer
			// we can't call process_substream_payload() yet, because the
			// substream_id byte is actually the 1st byte of payload

			j = s_fwrite(&substream_id, 1, 1, m_outfp) ;
			if ( j != 1)
			{
				sprintf( m_szText, "process_substream() : Error writing to output file %s.", m_bs.filename );
				TextLogErr( DMXERR_ENDOFSTREAM, m_szText );
				return DMXERR_ENDOFSTREAM;
				//exit (1);
			}
			cbPayload += j;
		}
		s.substream_bytes[substream_id] += 1;

		j = process_substream_payload( stream_id, substream_id, 3 );
		cbPayload += j;
//		for (j = 0; j < 3; j++)
//		{
//		  k = bs_getbits(&m_bs, 8);
//		  if (s_fwrite(&k, 1, 1, m_outfp) != 1)
//		  {
//			printf("Error writing to output file %s.", argv[3]);
//			exit (1);
//		  }
//		}
	} // end if ( substream_id ... )

	bytes_to_get = s.PES_packet_length - s.PES_header_data_length - 7;
	j = process_substream_payload( stream_id, substream_id, bytes_to_get );
	cbPayload += j;
	if ( fAC3data )	// is there *any* AC3 data (0x80 through 0x8F) ?
		s.ac3_bytes += j;

//		for (j = 0; j < s.PES_packet_length - s.PES_header_data_length - 7; j++)
//		{
//		  k = bs_getbits(&m_bs, 8);
//		  if (s_fwrite(&k, 1, 1, m_outfp) != 1)
//		  {
//			printf("Error writing to output file %s.", argv[3]);
//			exit (1);
//		  }
//		}
//	  } // end if (substream_id == substream_to_get)
//	  else
//		for (j = 0; j < s.PES_packet_length - s.PES_header_data_length - 4; j++)
//		  bs_getbits(&m_bs, 8);

/* moved back to go_demux()

	} // end if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag) )
	else
	{
	  // nope, this isn't a substream, it's just regular stream
	  if ( stream_id == stream_to_get )
	  {
		j = s_fwrite(&substream_id, 1, 1, m_outfp);
	    if ( j != 1)
		{
		  fprintf(stderr,"Error writing to output file %s.", m_bs.filename );
		  exit (1);
		}
		cbPayload += j;
	  }

	  bytes_to_get = s.PES_packet_length - s.PES_header_data_length - 4;
	  j = process_stream_payload( stream_id, bytes_to_get );
	  cbPayload += j;
	*/

//	  for (j = 0; j < s.PES_packet_length - s.PES_header_data_length - 4; j++)
//	  {
//		k = bs_getbits(&m_bs, 8);
//		if (s_fwrite(&k, 1, 1, m_outfp) != 1)
//		{
//		  printf("Error writing to output file %s.", argv[3]);
//		  exit (1);
//		}
//	  }

//	} // end if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag) && (substream_to_get > STREAM_NOT_SET))

//} // end if (stream_id == stream_to_get)
//else
//	for (j = 0; j < s.PES_packet_length - s.PES_header_data_length - 4; j++)
//	  bs_getbits(&m_bs, 8);


//			  s.stream_bytes[stream_id] += s.PES_packet_length - s.PES_header_data_length - 3;
//			  if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag))
//			  {
//				s.substream_packets[substream_id]++;
//				s.substream_bytes[substream_id] += s.PES_packet_length - s.PES_header_data_length - 3;
//				if (substream_id == SUBSTREAM_AC3)
//				  s.ac3_bytes += s.PES_packet_length - s.PES_header_data_length - 7;
//			  }

	return cbPayload;

}	// cDemuxer::process_substream()


// reset the demultiplexor, will not reset if a file is still open
DWORD
cDemuxer::reset_demux( void )
{
	unsigned long i;

	TextLogMsg( LOG_NEW_ENTRY, "reset_demux() : " );

	if ( BSOPEN( &m_bs ) )
	{
		sprintf( m_szText, "reset_demux() : '%s' still open!", 
			m_bs.filename );
		TextLogErr( DMXERR_FAIL, m_szText );
		return DMXERR_FAIL;
	}

	for (i = 0; i < 256; i++)
	{
		s.streams_found[i] = false;
		s.stream_packets[i] = 0;
		s.stream_bytes[i] = 0;
		s.substreams_found[i] = false;
		s.stream_encrypted[i] = false;
		s.substream_encrypted[i] = false;
		s.substream_packets[i] = 0;
		s.substream_bytes[i] = 0;
	}
	s.ac3_bytes = 0;
	s.pack_packets = 0;
	s.system_packets = 0;
	m_outfp = NULL;
	s.mpeg2_flag = false;
	s.firsttime = true;
	s.dwLastHeaderLBA = -1;  // initial value, -1 is invalid
	s.dwCurrHeaderLBA = -1;

	// reset bitstream vars
	m_bs.dwLBAPos = 0;
	m_bs.dwBytePos = 0;
	m_bs.dwLBAOffset = 0;
	m_bs.buf = NULL;


	stream_to_get = STREAM_NOT_SET;
	substream_to_get = STREAM_NOT_SET;
	m_cbFwriteBuf = 0;	// counter, #bytes stored in m_cbFwriteBuf
	m_pFwriteBuf = NULL;		// point to FwriteBuffer

	return DMXERR_OK;
}  // cDemuxer::reset_demux()


// go_demux() is the *public* interface.
// An application initializes the cDemux object with a filename, and 
// the desired streams to rip (stream_to_get, substream_to_get), then finally
// calls go_demux().
//
// go_demux() does the rest :)
//
//  some organizational changes from Brent Beyeler's BBDMUX 1.2.
//  cDemuxer's go_demux() [ used to be BBDMUX's main() ] locates all streams,
//  and sends all streams to process_stream() or process_substream().
//
//  The filtering-process (extracting the requested stream) occurs in these
//  subfunctions.
//
//  The subfunctions are expected to return the # bytes
//  extracted from the REQUESTED streams.  go_demux() also directly writes
//  a few bytes to the output-buffer  m_pFwriteBuf[].  If any changes are
//  made to the function, care must be taken to ensure cbPayload is updated
//  correctly!  At the end of the function,
//    (*pcbBuffer) = cbPayload;
//
//    Thus, cbPayload informs the calling-function the #payload-bytes in 
//     the output *buffer!
 
//  since go_demux() directly 
int 
cDemuxer::go_demux( unsigned char *buffer, unsigned long *pcbBuffer )
{
  //Bit_stream_struc m_bs;
  unsigned long i, j;
  unsigned long bytes_to_get = 0;
  unsigned long cbPayload = 0; // counter, bytes extracted from DESIRED stream
	// cbPayload is incremented from process_stream() and process_substream()'s
    // return values.  Those functions are expected to return the # bytes
    // extracted from the DESIRED stream.  In other words, if go_demux() calls
    // either with the wrong stream_id, they should return 0 (to indicate that
    // no payload bytes were read.)

  /*
  unsigned long s.PES_packet_length, s.PES_header_data_length;
  unsigned long s.PTS_DTS_flags, s.ESCR_flag, s.ES_rate_flag, s.DSM_trick_mode_flag;
  unsigned long s.additional_copy_info_flag, s.PES_CRC_flag, s.PES_extension_flag; 
  unsigned long s.PES_private_data_flag, s.pack_header_field_flag;
  unsigned long s.program_packet_sequence_counter_flag, s.PSTD_buffer_flag;
  unsigned long s.PES_extension_flag_2, s.PES_extension_field_length;
  unsigned long s.bytes_used, s.encrypted;
  int substream_to_get, stream_to_get;
  */

  unsigned char stream_id = 0, substream_id= 0;
  
  /*
  static bool s.s.firsttime = true;
  unsigned long s.stream_packets[256], s.stream_bytes[256], s.pack_packets;
  unsigned long s.system_packets, s.ac3_bytes;
  unsigned long s.substream_packets[256], s.substream_bytes[256];
  */
  
  FILE *m_outfp = NULL; // dummy variable for s_fwrite()
  /*
  bool s.streams_found[256];
  bool s.stream_encrypted[256];
  bool s.substream_encrypted[256];
  bool s.substreams_found[256];
  bool s.vob_flag, s.mpeg2_flag;
  */
  //char tmpStr[ TEMP_STRLENGTH ];
  bool f_ueof = false;	// unexpected end-of-bitstream
		// this flag is set if EOF is encountered in the middle of a data pack
  

  // moved to reset_demux();

  /*
  for (i = 0; i < 256; i++)
  {
	s.streams_found[i] = false;
	s.stream_packets[i] = 0;
	s.stream_bytes[i] = 0;
	s.substreams_found[i] = false;
	s.stream_encrypted[i] = false;
	s.substream_encrypted[i] = false;
	s.substream_packets[i] = 0;
	s.substream_bytes[i] = 0;
  }
  s.ac3_bytes = 0;
  s.pack_packets = 0;
  s.system_packets = 0;
  m_outfp = NULL;
  stream_to_get = STREAM_NOT_SET;
  substream_to_get = STREAM_NOT_SET;
  s.mpeg2_flag = false;
  */

  /*
  printf("bbDMUX - version 1.2, by Brent Beyeler (beyeler@home.com)\n\n");
  if (argc < 2 || argc == 3)
  {
	printf("bbDUMX is an MPEG program stream de-multiplexor\n\n");
	printf("Usage: bbDMUX  MPEG filename  <stream id  saved name>  <substream id>\n\n");
	printf("Examples:\n");
	printf("  To get a list of the stream and substream id's present in an MPEG file\n");
	printf("    bbDMUX  testfile.mpg\n\n");
	printf("  To save a video stream (id 0xe0) to the file video.m2v\n");
	printf("    bbDMUX  testfile.m2p  0xe0  video.m2v\n\n");
	printf("  To save an audio stream (id 0xc0) to the file audio.m2a\n");
	printf("    bbDMUX  testfile.mpg  0xc0  audio.m2a\n\n");
	printf("  To save a private 1 stream (id 0xbd) to the file private1.pvt\n");
	printf("    bbDMUX  testfile.m2p  0xbd  private1.pvt\n\n");
	printf("  To save an AC3 audio stream (substream 0x80) to the file audio.ac3\n");
	printf("    bbDMUX  testfile.vob  0xbd  audio.ac3  0x80\n");
	 exit (1);
  }
  */


/*
  if (argc > 2)
  {
	sscanf(argv[2], "0x%x", &i);
	if (i < 0 || i > 0xFF)
	{
	  printf("Stream ID must be in the range 0..0xFF\n");
	  exit (1);
	}
	stream_to_get = i;
	if (argc > 3)
	{
	  m_outfp = fopen(argv[3], "wb");
	  if (!m_outfp)
	  {
		printf("Unable to open %s for output\n", argv[3]);
		exit (1);
	  }
	}
	else
	{
	  printf("An output filename must be specified\n");
	  exit(1);
	}

	if (argc > 4)
	{
	  sscanf(argv[4], "0x%x", &i);
	  if (i < 0 || i > 0xFF)
	  {
		printf("Substream ID must be in the range 0..0xFF\n");
		exit (1);
	  }
	  substream_to_get = i;
	  s.vob_flag = true;
	}
  }
  else
	printf("Scanning for stream id's, press control-c to quit ...\n");

  if (m_outfp)
	if ((s.vob_flag) && (stream_to_get == PRIVATE_STREAM_1) && (substream_to_get > STREAM_NOT_SET))
	  printf("Saving stream 0x%02X, substream 0x%02X into file %s...\n", stream_to_get, substream_to_get, argv[3]);
	else
	  printf("Saving stream 0x%02X into file %s...\n", stream_to_get, argv[3]);
*/

  if ( !BSOPEN( &m_bs ) )
  {
	TextLogErr( DMXERR_FAIL, "go_demux() : no bitstream is open!" );
	return DMXERR_FAIL;
  }

  // validate the input pointers
  if ( buffer == NULL || pcbBuffer == NULL )
  {
	TextLogErr( DMXERR_FAIL, "go_demux() : input pointers == NULL!" );
	if ( pcbBuffer != NULL )
		(*pcbBuffer) = 0;  // 0 bytes demuxed due to error
	return DMXERR_FAIL;
  }

	// initialize the local temporary buffer with the supplied buffer-space
	// s_fwrite() needs these two member vars to function.
	// s_fwrite() writes to m_pFwriteBuf[], and updates m_cbFwriteBuf
  m_pFwriteBuf= buffer;  // assume buffer is valid for duration of this call!
  m_cbFwriteBuf = 0; // reset counter (#bytes written) to 0

// ----------- main demux loop 
  do
  {
	i = bs_getbits(&m_bs, 32);
nextpass:
	if (s.firsttime)
	{
	  printf("\ncDemuxer::go_demux() : File %s ", m_bs.filename );
	  if (i != PACK_START_CODE)
	  {
		printf("is not an MPEG Program Stream\n" );
		return DMXERR_ENDOFSTREAM;
	  }
	  s.mpeg2_flag = bs_look_ahead(&m_bs, 2) == 1; // check if MPEG-1 or MPEG-2
	  if (s.mpeg2_flag)
		printf("is a MPEG-2 Program Stream\n" );
	  else
		printf("is a MPEG Program Stream\n");
	  s.firsttime = false;
	} // endif ( s.firsttime )
	if ( bs_end_bs( &m_bs ) )	// if end of bitstream
		f_ueof = TRUE; // we've reached end of file
	else
	switch (i)
	{
	  case PACK_START_CODE:
		s.pack_packets++;
		if (s.mpeg2_flag)
		{
//		  printf("\nMPEG-2 PACK START @ LBA:%ld, Byte Offset:%ld\n",
//			m_bs.dwLBAPos, m_bs.dwBytePos );

		  bs_getbits(&m_bs, 32);   // MPEG2 pack header
		  bs_getbits(&m_bs, 32);
		  bs_getbits(&m_bs, 13);
		  i = bs_getbits(&m_bs, 3);
		  for (j = 0; j < i; j++)
			bs_getbits(&m_bs, 8);  // stuffing bytes
		}
		else
		{
//		  printf("\nMPEG-1 PACK START @ LBA:%ld, Byte Offset:%ld\n",
//			m_bs.dwLBAPos, m_bs.dwBytePos );
		  bs_getbits(&m_bs, 32);   // MPEG1 pack header
		  bs_getbits(&m_bs, 32);
		}
		break;

	  case SYSTEM_HEADER_START_CODE:
		s.system_packets++;
//		printf("\nSystem_header_start_code @ LBA:%ld, Byte Offset:%ld\n",
//			m_bs.dwLBAPos, m_bs.dwBytePos );

		// record LBA address of this SYSTEM_HEADER_PACKET
		s.dwLastHeaderLBA = s.dwCurrHeaderLBA;  // update last LBA
		bs_get_lba( &m_bs, &(s.dwCurrHeaderLBA) ); // update current LBA

		bs_getbits(&m_bs, 32);	 // system header
		bs_getbits(&m_bs, 32);
		while (bs_look_ahead(&m_bs, 1) == 1)
		  bs_getbits(&m_bs, 24); // P-STD info
		break;

	  case MPEG_PROGRAM_END_CODE:

		break;

	  default:
		if ((i >> 8) != PACKET_START_CODE_PREFIX)
		{
#ifdef SHOW_RESYNCHS
		  if (argc > 3)
		  {
			printf("\nUnknown bits in stream = 0x%08X at byte %.0f\n", i, bs_sstell(&m_bs) / 8 - 4);
			printf("\nAttempting resynch ... ");
		  }
#endif
		  if (bs_seek_sync(&m_bs, PACKET_START_CODE_PREFIX, 24))
		  {
			i = 0x00000100 | bs_getbits(&m_bs, 8);
#ifdef SHOW_RESYNCHS
			if (argc > 3)
			  printf("succesfully resynched\n");
#endif
			sprintf( m_szText, "go_demux() : resynch successful...now @ lba<%d>",
				m_bs.dwLBAPos );
			TextLogMsg( LOG_NEW_ENTRY, m_szText );
			printf( "%s\n", m_szText );
			goto nextpass;
		  }
#ifdef SHOW_RESYNCHS
		  if (argc > 3)
			printf("could not resynch\n");
		  else
#endif
			sprintf( m_szText, "go_demux() : Unknown bits in stream = 0x%08X at byte %.0f\n", 
				i, bs_sstell(&m_bs) / 8 - 4);
			TextLogErr( DMXERR_ENDOFSTREAM, m_szText );
			return DMXERR_ENDOFSTREAM;
		} // end if ((i >> 8) != PACKET_START_CODE_PREFIX)
		stream_id = i & 0x000000FF;
		s.stream_packets[stream_id]++;

		s.PES_packet_length = bs_getbits(&m_bs, 16);
		s.encrypted = 0;
		switch (stream_id)
		{
		  case PROGRAM_STREAM_MAP:
		  case ECM_STREAM:
		  case EMM_STREAM:
		  case PROGRAM_STREAM_DIRECTORY:
		  case DSMCC_STREAM:
		  case ITUTRECH222TYPEE_STREAM:
		  case PRIVATE_STREAM_2:
		  case PADDING_STREAM:
			bytes_to_get = s.PES_packet_length;
			j = process_stream_payload( stream_id, bytes_to_get );
			// process_stream_payload updates stream_bytes[]
			cbPayload += j;
//			if ( j != bytes_to_get )
//			{
//				f_ueof = true;  // hit end of bitstream in mid-packet!
//			}
			break;

		  default:
			s.bytes_used = 0;
			if (!s.mpeg2_flag)
			{
			  /* flush stuffing bytes */
			  while (bs_look_ahead(&m_bs, 8) == 0xff)
			  {
				bs_getbits(&m_bs, 8);
				s.bytes_used++;
			  }

			  if (bs_look_ahead(&m_bs, 2) == 1)
			  {
				/* Get STD buffer size */
				bs_getbits(&m_bs, 16);
				s.bytes_used += 2;
			  }

			  if (bs_look_ahead(&m_bs, 4) == 2)
			  {
				/* Get presentation time stamp and flag it as present in packet*/
				bs_getbits(&m_bs, 32);
				bs_getbits(&m_bs, 8);
				s.bytes_used += 5;
			  }
			  else
				if (bs_look_ahead(&m_bs, 4) == 3)
				{
				  /* Get presentation time stamp and decoding time stamp */
				  bs_getbits(&m_bs, 32);
				  bs_getbits(&m_bs, 8);
				  bs_getbits(&m_bs, 32);
				  bs_getbits(&m_bs, 8);
				  s.bytes_used += 10;
				}
				else
				{
				  if (bs_getbits(&m_bs, 8) != 0x0f)
				  {
					printf("Invalid bits in MPEG-1 file %s.", m_bs.filename);
					return DMXERR_ENDOFSTREAM;
					//exit (1);
				  }
				  s.bytes_used++;
				} // end if (bs_look_ahead(&m_bs, 4) == 3)
			  bytes_to_get = s.PES_packet_length - s.bytes_used ;
			  j = process_stream_payload( stream_id, bytes_to_get );
			  cbPayload += j;
//			  if ( j != bytes_to_get )
//			  {
//				  f_ueof = true;  // hit end of bitstream in mid-packet!
//			  }
//			  s.stream_bytes[stream_id] += s.PES_packet_length - s.bytes_used;
//			  if (stream_id == stream_to_get)
//			  {
//				for (j = 0; j < s.PES_packet_length - s.bytes_used; j++)
//				{
//				  k = bs_getbits(&m_bs, 8);
//				  if (s_fwrite(&k, 1, 1, m_outfp) != 1)
//				  {
//					printf("Error writing to output file %s.", argv[3]);
//					exit (1);
//				  }
//				}
//			  }
//			  else
//				for (j = 0; j < s.PES_packet_length - s.bytes_used; j++)
//				  bs_getbits(&m_bs, 8);
			} // if (!s.mpeg2_flag)
			else
			{
			  bs_getbits(&m_bs, 2);
			  s.encrypted = bs_getbits(&m_bs, 2);	// PES_scrambling_control
			  bs_getbits(&m_bs, 4);
			  s.PTS_DTS_flags = bs_getbits(&m_bs, 2);
			  s.ESCR_flag = bs_get1bit(&m_bs);
			  s.ES_rate_flag = bs_get1bit(&m_bs);
			  s.DSM_trick_mode_flag = bs_get1bit(&m_bs);
			  s.additional_copy_info_flag = bs_get1bit(&m_bs);
			  s.PES_CRC_flag = bs_get1bit(&m_bs);
			  s.PES_extension_flag = bs_get1bit(&m_bs);
			  s.PES_header_data_length = bs_getbits(&m_bs, 8);
			  if (s.PTS_DTS_flags > 1)
			  {
				bs_getbits(&m_bs, 32);	 // PTS stamp
				bs_getbits(&m_bs, 8);
				s.bytes_used += 5;
			  }
			  if (s.PTS_DTS_flags > 2)
			  {
				bs_getbits(&m_bs, 32);	 // DTS stamp
				bs_getbits(&m_bs, 8);
				s.bytes_used += 5;
			  }

			  if (s.ESCR_flag == 1)
			  {
				bs_getbits(&m_bs, 32);
				bs_getbits(&m_bs, 16);
				s.bytes_used += 6;
			  }

			  if (s.ES_rate_flag == 1)
			  {
				bs_getbits(&m_bs, 24);
				s.bytes_used += 3;
			  }

			  if (s.DSM_trick_mode_flag == 1)
			  {
				bs_getbits(&m_bs, 8);
				s.bytes_used++;
			  }

			  if (s.additional_copy_info_flag == 1)
			  {
				bs_getbits(&m_bs, 8);
				s.bytes_used++;
			  }

			  if (s.PES_CRC_flag == 1)
			  {
				bs_getbits(&m_bs, 16);
				s.bytes_used += 2;
			  }

			  if (s.PES_extension_flag == 1)
			  {
				s.PES_private_data_flag = bs_get1bit(&m_bs);
				s.pack_header_field_flag = bs_get1bit(&m_bs);
				s.program_packet_sequence_counter_flag = bs_get1bit(&m_bs);
				s.PSTD_buffer_flag = bs_get1bit(&m_bs);
				bs_getbits(&m_bs, 3);
				s.PES_extension_flag_2 = bs_get1bit(&m_bs);
				s.bytes_used++;

				if (s.PES_private_data_flag == 1)
				{
				  for (j = 0; j < 128; j++)
					bs_getbits(&m_bs, 8);
				  s.bytes_used += 128;
				}

				if (s.pack_header_field_flag == 1)
				{
				  printf("    pack header field flag value not allowed in program streams\n");
				  return DMXERR_ENDOFSTREAM;
				  //exit(1);
				}

				if (s.program_packet_sequence_counter_flag == 1)
				{
				  bs_getbits(&m_bs, 16);
				  s.bytes_used += 2;
				}

				if (s.PSTD_buffer_flag == 1)
				{
				  bs_getbits(&m_bs, 16);
				  s.bytes_used += 2;
				}

				if (s.PES_extension_flag_2 == 1)
				{
				  bs_get1bit(&m_bs);
				  s.PES_extension_field_length = bs_getbits(&m_bs, 7);
				  s.bytes_used++;
				  for (j = 0; j < s.PES_extension_field_length; j++)
				  {
					bs_getbits(&m_bs, 8);
					s.bytes_used++;
				  }
				}
			  } // end if (s.PES_extension_flag == 1)
			  for (j = 0; j < s.PES_header_data_length - s.bytes_used; j++)
				bs_getbits(&m_bs, 8);

			  substream_id = bs_getbits(&m_bs, 8);

// these housekeeping updates are now handled by process_substream() &
// and its subfunction, process_substream_payload()
//			  s.stream_bytes[stream_id] += s.PES_packet_length - s.PES_header_data_length - 3;
//			  if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag))
//			  {
//				s.substream_packets[substream_id]++;
//				s.substream_bytes[substream_id] += s.PES_packet_length - s.PES_header_data_length - 3;
//				if (substream_id == SUBSTREAM_AC3)
//				  s.ac3_bytes += s.PES_packet_length - s.PES_header_data_length - 7;
//			  }

			  // check if the passage is a SUBSTREAM or just a regular STREAM
			  if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag) && 
				substream_to_get != STREAM_NOT_SET )
			  { // yes, this is a SUBSTREAM
				j = process_substream( stream_id, substream_id );
				cbPayload += j;
			  }
			  else
			  { 
				// nope, this isn't a substream, it's just regular stream
				if ( stream_id == stream_to_get )
				{
				// manually write this byte to the output-buffer
				// we can't call process_stream_payload() yet, because the
				// substream_id byte is actually the 1st byte of payload

				  j = s_fwrite(&substream_id, 1, 1, m_outfp);
				  if ( j != 1)
				  {
					sprintf( m_szText, "go_demux() : Error writing to output file %s.", m_bs.filename );
					TextLogErr( DMXERR_ENDOFSTREAM, m_szText );
					return DMXERR_ENDOFSTREAM;
					//exit (1);
				  }
				  cbPayload += j;
				} // end if ( stream_id == stream_to_get )

				s.stream_bytes[stream_id] += 1;

				bytes_to_get = s.PES_packet_length - s.PES_header_data_length - 4;
				j = process_stream_payload( stream_id, bytes_to_get );
				cbPayload += j;

			  } // end if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag) ..
			} // end else of if (!s.mpeg2_flag)

			break;
		} // endswitch (stream_id)


	/*
		if (argc == 2)
		{
		  if ((!s.streams_found[stream_id]) || ((s.vob_flag) && (stream_id == PRIVATE_STREAM_1) && (!s.substreams_found[substream_id])))
		  {
			switch (stream_id)
			{
			  case PROGRAM_STREAM_MAP:
				printf("Found stream id 0x%02X  = Program Stream Map\n", stream_id);
				break;
			  case PRIVATE_STREAM_1:
				if (s.vob_flag)
				{
				  switch (substream_id)
				  {
					case SUBSTREAM_AC3:
					  printf("Found stream id 0x%02X  = Private Stream 1, substream 0x80, AC3 audio\n", stream_id);
					  break;
					default:
					  printf("Found stream id 0x%02X  = Private Stream 1, substream 0x%02X\n", stream_id, substream_id);
					  break;
				  }
				  s.substreams_found[substream_id] = true;
				}
				else
				  printf("Found stream id 0x%02X  = Private Stream 1\n", stream_id);
				break;
			  case PRIVATE_STREAM_2:
				printf("Found stream id 0x%02X  = Private Stream 2\n", stream_id);
				break;
			  case ECM_STREAM:
				printf("Found stream id 0x%02X  = ECM Stream\n", stream_id);
				break;
			  case EMM_STREAM:
				printf("Found stream id 0x%02X  = EMM Stream\n", stream_id);
				break;
			  case PROGRAM_STREAM_DIRECTORY:
				printf("Found stream id 0x%02X  = Program Stream Directory\n", stream_id);
				break;
			  case DSMCC_STREAM:
				printf("Found stream id 0x%02X  = DSMCC Stream\n", stream_id);
				break;
			  case ITUTRECH222TYPEE_STREAM:
				printf("Found stream id 0x%02X  = ITU-T Rec. H.222.1 type E\n", stream_id);
				break;
			  case PADDING_STREAM:
				printf("Found stream id 0x%02X  = Padding Stream\n", stream_id);
				break;
			  default:
				if ((stream_id >= 0xE0) && (stream_id <= 0xEF))
				  printf("Found stream id 0x%02X  = Video Stream %d\n", stream_id, stream_id - 0xE0);
				else
				  if ((stream_id >= 0xC0) && (stream_id <= 0xDF))
					printf("Found stream id 0x%02X  = Audio Stream %d\n", stream_id, stream_id - 0xC0);
				  else
					printf("Found stream id 0x%02X  = other stream\n", stream_id);
			}
			s.streams_found[stream_id] = true;
		  }
		  if ((stream_id == PRIVATE_STREAM_1) && (s.vob_flag))
		  {
			if (s.encrypted)
			{
			  if (!s.substream_encrypted[substream_id])
			  {
				printf("Stream id 0x%02X, substream 0x%02X is s.encrypted\n", stream_id, substream_id);
				s.substream_encrypted[substream_id] = true;
			  }
			}
		  }
		  else
		  {
			if (s.encrypted)
			{
			  if (!s.stream_encrypted[stream_id])
			  {
				printf("Stream id 0x%02X is s.encrypted\n", stream_id);
				s.stream_encrypted[stream_id] = true;
			  }
			}
		  }
		}
	*/
	} // end switch( i )
  // stay in loop until we get at least one payload byte, then quit
  } while ( (i != MPEG_PROGRAM_END_CODE) && (!f_ueof) 
	  && (cbPayload == 0) ); // && (!feof(m_bs.pt)));

// ----- end of main do() demux-loop

	// set pcbBuffer = #bytes payload we retrieved
	(*pcbBuffer ) = cbPayload;
  
	//printf("\ncbPayload = %ld, m_cbFwriteBuf = %ld", cbPayload, m_cbFwriteBuf );
	
	if ( i ==MPEG_PROGRAM_END_CODE )
	{
		//bs_update_lba( &m_bs );
		printf("\ngo_demux() : MPEG_PROGRAM_END_CODE @ LBA:%d\n", m_bs.dwLBAPos );
	}
	if ( f_ueof )
		printf("\nf_ueof!\n");

	// reset the variables
	m_pFwriteBuf = NULL;
	m_cbFwriteBuf = 0; 

	if ( f_ueof || i==MPEG_PROGRAM_END_CODE || cbPayload == 0 )
		return DMXERR_ENDOFSTREAM;
	else
		return DMXERR_OK;
}


void
cDemuxer::print_summary( void )
{
	unsigned long i, j;

	TextLogMsg( LOG_NEW_ENTRY, "print_summary() :" );

  printf("\n\tBBDMUX v1.3 Summary:\n\t--------------------");
  if (s.mpeg2_flag)
	printf("\n\tMPEG2 Packs = %d, PES packet length (most recent) = %ld", s.pack_packets,
		s.PES_packet_length );
  else
	printf("\n\tMPEG1 Packs = %d", s.pack_packets);
  if (s.system_packets)
	printf("\n\tSystem headers = %d  (LBA of last 2 headers = #%ld , #%ld )\n\t", 
	s.system_packets, s.dwCurrHeaderLBA, s.dwLastHeaderLBA );
  
  for (i = 0; i < 256; i++)
  {
	if (s.stream_packets[i])
	{
	  switch (i)
	  {
		case PROGRAM_STREAM_MAP:
		  printf("Program Stream Map packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case PRIVATE_STREAM_1:
		  printf("Private Stream 1 packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  if (s.vob_flag)
		  {
			for (j = 0; j < 256; j++)
			{
			  if (s.substream_packets[j])
			  {
				switch (j)
				{
			//	  case SUBSTREAM_AC3:
			//		printf("    AC3 Audio substream packets = %d, total bytes = %d\n", 
			//			s.substream_packets[j], s.ac3_bytes);
			//		break;
				  default:
					printf("    Substream 0x%02X packets = %d, total bytes = %d\n", 
						j, s.substream_packets[j], s.substream_bytes[j]);
					break;
				}
			  }
			}
		  }
		  break;
		case PRIVATE_STREAM_2:
		  printf("Private Stream 2 packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case ECM_STREAM:
		  printf("ECM Stream packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case EMM_STREAM:
		  printf("EMM Stream packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case PROGRAM_STREAM_DIRECTORY:
		  printf("Program Stream Directory packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case DSMCC_STREAM:
		  printf("DSMCC Stream packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case ITUTRECH222TYPEE_STREAM:
		  printf("ITU-T Rec. H.222.1 type E packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		case PADDING_STREAM:
		  printf("Padding Stream packets = %d, total bytes = %d\n\t", s.stream_packets[i], s.stream_bytes[i]);
		  break;
		default:
		  if ((i >= 0xE0) && (i <= 0xEF))
			printf("Video stream %d packets = %d, total bytes = %d\n\t", i - 0xE0, s.stream_packets[i], s.stream_bytes[i]);
		  else
			if ((i >= 0xC0) && (i <= 0xDF))
			  printf("Audio stream %d packets = %d, total bytes = %d\n\t", i - 0xC0, s.stream_packets[i], s.stream_bytes[i]);
			else
			  printf("Other stream 0x%02X packets = %d, total bytes = %d\n\t", i, s.stream_packets[i], s.stream_bytes[i]);
	  }
	}
  }
  //bs_close_bit_stream_r(&m_bs);
  if (m_outfp)
	fclose(m_outfp);
  return;
}  // cDemuxer::print_summary( void )



// set filename, automatically opens the input_file
int 
cDemuxer::set_filename( char *input_filename )
{
	char tmpStr[ TEMPSTR_LENGTH ];
	long bsrc = 0; // bs_open_bit_streamr() return code

	sprintf( m_szText, "set_filename( '%s' ) : ", input_filename );
	TextLogMsg( LOG_NEW_ENTRY, m_szText );

	// first check if a bitstream is already open
	if ( BSOPEN( (&m_bs) ) )
	{
		TextLogErr( DMXERR_OPEN, "set_filename() : a bitstream is already open." );
		return DMXERR_OPEN;
	}

	if ( stream_to_get == STREAM_NOT_SET )
	{
		TextLogErr( DMXERR_FAIL, "set_filename() : no target stream_id selected. use set_target_id()!");
		return DMXERR_FAIL;
	}


	strcpy(tmpStr, m_bs.filename);
	strlwr(tmpStr);
	strcpy( m_bs.filename, input_filename );

	//s.vob_flag = (strstr(tmpStr, ".vob") != NULL);
	s.vob_flag = true;
	bsrc = bs_open_bit_stream_r (&m_bs, DEMUX_BUFFER_SIZE);
	printf("\ncDemuxer::set_filename( %s, STREAM_ID == 0x%02X );",
		m_bs.filename, stream_to_get );

	return bsrc;	// bsrc indicates whether operation succeeded or failed
}

int 
cDemuxer::set_target_id( unsigned char stream_id, unsigned char substream_id )
{
	sprintf( m_szText, "set_target_id( ID(%02X), SUB_ID(%02X) ) : ", 
		stream_id, substream_id );
	TextLogMsg( LOG_NEW_ENTRY, m_szText );

	// first check if a bitstream is already open
	if ( BSOPEN( (&m_bs) ) )
	{
		TextLogErr( DMXERR_OPEN, "set_target_id() : a bitstream is already open." );
		return DMXERR_OPEN;
	}

	if ( stream_id == STREAM_NOT_SET )
	{
		sprintf( m_szText, "set_target_id() : %d is not a valid stream_to_get",
			stream_id );
		TextLogErr( DMXERR_FAIL, m_szText );
		return DMXERR_FAIL;
	}

	stream_to_get = stream_id;
	substream_to_get = substream_id ;
	return 0;
}

// seek to logical block address
// does not check if LBA is past the end of current file!
int
cDemuxer::fseek_lba( DWORD dwLBA )
{
	return bs_fseek_lba( &m_bs, dwLBA );
}


// returns the LBA of the most recently decoded SYSTEM_HEADER,
// or (-1) FAIL
DWORD 
cDemuxer::get_system_llba( void )
{
	// if file isn't open, get_system_llba() should return -1
	if ( BSOPEN( (&m_bs) ) )
		return s.dwCurrHeaderLBA;	// bitstream is open
	else
	{	// fail, no bitstream is open
		TextLogErr( DMXERR_FAIL, "get_system_llba() : no bitstream is open.");
		return -1;	
	}
}



// Start at LBA (dwStart.)  Scan for SYSTEM_HEADER.  Stops at the first
// valid SYSTEM_HEADER, and reseeks to the beginning of that LBA.
// returns BSERR_OK if successful, BSERR_FAIL otherwise.
DWORD
cDemuxer::lba_seek_next_sysh( DWORD dwStart )
{
  unsigned long i, j; // working vars
  unsigned dwCurrentLBA = dwStart;
  BOOL fGotHeader = FALSE;	// loop status flag

  sprintf( m_szText, "lba_seek_next_sysh( dwStart==%lu ) : ", dwStart );
  TextLogMsg( LOG_NEW_ENTRY, m_szText );

  i = fseek_lba( dwCurrentLBA );
  if ( i != BSERR_OK )
  {
	TextLogErr( DMXERR_FAIL, "lba_seek_next_sysh() : fseek_lba() failed!" );
	return DMXERR_FAIL;
  }
  printf ("\nlba_seek_next_sysh() : Scanning LBA#%lu ...", dwCurrentLBA );

  s.mpeg2_flag = TRUE;
  do
  {
    i = bs_getbits( &m_bs, 32);	// get 32 bits

	switch (i)
	{
	  case PACK_START_CODE:
		if (s.mpeg2_flag)
		{
//		  printf("\nMPEG-2 PACK START @ LBA:%ld, Byte Offset:%ld\n",
//			m_bs.dwLBAPos, m_bs.dwBytePos );

		  bs_getbits(&m_bs, 32);   // MPEG2 pack header
		  bs_getbits(&m_bs, 32);
		  bs_getbits(&m_bs, 13);
		  i = bs_getbits(&m_bs, 3);
		  for (j = 0; j < i; j++)
			bs_getbits(&m_bs, 8);  // stuffing bytes
		}
		else
		{
//		  printf("\nMPEG-1 PACK START @ LBA:%ld, Byte Offset:%ld\n",
//			m_bs.dwLBAPos, m_bs.dwBytePos );
		  bs_getbits(&m_bs, 32);   // MPEG1 pack header
		  bs_getbits(&m_bs, 32);
		}
		break;

	  case SYSTEM_HEADER_START_CODE:
		fGotHeader = TRUE;
		// record LBA address of this SYSTEM_HEADER_PACKET
		s.dwLastHeaderLBA = s.dwCurrHeaderLBA;  // update last LBA
		bs_get_lba( &m_bs, &(s.dwCurrHeaderLBA) ); // update current LBA

		bs_getbits(&m_bs, 32);	 // system header
		bs_getbits(&m_bs, 32);
		while (bs_look_ahead(&m_bs, 1) == 1)
		  bs_getbits(&m_bs, 24); // P-STD info
		break;

	  case MPEG_PROGRAM_END_CODE:
		  printf("\rlba_seek_next_sysh() : LBA#%ld ... MPEG_PROGRAM_END_CODE DOH!!!\n",  
				dwCurrentLBA );
		break;

	  default:
		++dwCurrentLBA;	// increment LBA index
		i = fseek_lba( dwCurrentLBA ); // seek the next LBA
		if ( i != BSERR_OK )
		{
			TextLogErr( BSERR_FAIL, "lba_seek_next_sysh() : fseek_lba() failed!" );
			return BSERR_FAIL;
		}
		if ( (dwCurrentLBA & 0x07) == 0 )  // only update console per 8 loops
			printf ("\rlba_seek_next_sysh() : Scanning LBA#%ld ...", 
				dwCurrentLBA );

	}  // end switch( i )

	// check for end of stream condition
	if ( m_bs.eobs )
	{
		TextLogErr( DMXERR_ENDOFSTREAM, "lba_seek_next_sysh() : end of stream!" );
		return DMXERR_ENDOFSTREAM;
	}
  } while ( fGotHeader == FALSE );

  // now reset the bitstream position to the start of the found LBA
  i = fseek_lba( dwCurrentLBA );
  if ( i != BSERR_OK )
  {
	TextLogErr( DMXERR_FAIL, "\nlba_seek_next_sysh() : fseek_lba() failed!" );
	return DMXERR_FAIL;
  }

  //bs_update_lba( &m_bs ); // update current position
  printf("\rlba_seek_next_sysh() : system_header_start_code @ LBA#%ld", 
	  dwCurrentLBA );

  return DMXERR_OK; // found SYSTEM_HEADER, success!
} // cDemuxer::lba_seek_next_sysh( DWORD dwStart )


//
// -----------------------------------------------------------------------------
//
// Housekeeping functions
//
// -----------------------------------------------------------------------------
//


void 
cDemuxer::TextLogErr( DWORD errorID, char text[] )
{
	char temp_text[ TEXTMSG_LENGTH ];
	static DWORD dwPurgeCount = 0;

	if ( strlen( m_szErrorLog ) > ( ( LOG_LENGTH >> 1 ) - TEXTMSG_LENGTH ) )
	{
		++dwPurgeCount;
		printf("\ncDemuxer::TextLogErr() : Warning log-size exceeded %ld bytes",
			strlen( m_szErrorLog ) );
		printf("\n\tPurging log...");
		sprintf( m_szErrorLog, "<ErrLog purged %ld times>\n", dwPurgeCount );
	}

	sprintf( temp_text, "\n#%ld cDemuxer::", m_logCounter );
	if ( text != NULL )
		strcat( temp_text, text );
	else
		strcat( temp_text, "<NULL>");
	strcat( m_szErrorLog, temp_text );
	strcpy( m_szLastLogErr, temp_text );

	printf( "%s", temp_text );


	sprintf( temp_text," Error ID = %0X ", errorID );
	strcat( m_szErrorLog, temp_text );
	strcat( m_szLastLogErr, temp_text );

	printf( "%s\n", temp_text );

	m_lastError = errorID; // update last error variable

	// counter is incremented by TextLogMsg()
} // cDemuxer::TextLogErr()


void 
cDemuxer::TextLogMsg( DWORD dwFlags, char text[] )
{
	char temp_text[ TEXTMSG_LENGTH ];

	static DWORD dwPurgeCount = 0;

	if ( strlen( m_szMessageLog ) > ( ( LOG_LENGTH >> 1 ) - TEXTMSG_LENGTH ) )
	{
		++dwPurgeCount;
		printf("\ncDemuxer::TextLogMsg() : Warning log-size exceeded %ld bytes",
			strlen( m_szMessageLog ) );
		printf("\n\tPurging log...");
		sprintf( m_szMessageLog, "<Log purged %ld times>\n", dwPurgeCount );
	}


	if ( dwFlags & LOG_NEW_ENTRY )
	{
		sprintf( temp_text, "\n#%ld cDemuxer::", m_logCounter );
		if ( text != NULL )
			strcat( temp_text, text );
		else
			strcat( temp_text, "<NULL>");
		strcat( m_szMessageLog, temp_text );

		m_lastError = BSERR_OK; // clear lastError flag
		strcpy( m_szLastLogErr, "<NO ERROR>"); // clear lasterror msg
		++m_logCounter;  // increment counter
	}
	else
	{
		strcat( m_szMessageLog, text );
	}
} // cDemuxer::TextLogMsg()
