//
// bbdmux12.h - header file for class cDemuxer
//
//  There are three files for class cDemuxer : 
//   bbdmux12.cpp, bbdmux12.h, bbdbits.cpp
//
// MPEG2AVI
// --------
//  v0.16B34 optimized bs_fseek_lba() for special case, bs_update_lba() 
//   modified when the (seek-target LBA) == (current LBA + 1), bs_fseek_lba 
//   uses bs_getbytes to "skip" to the next LBA, rather then using the 
//   slower my_seeki64() call.
//   thanks to the buffered read_ahead mechanism, the demux buffer_size
//   has been increased to 64KB
//  v0.16B33
//      updated the bbdmux portion (go_demux()) to v1.3, the demuxer properly
//      extracts AC3-substreams with ID : < 0x80 through 0x8F >
//      this issue never affected MPEG2AVI (since there's no audio support)
//  v0.16B32
//      tried to increase buffer size to 65536, but it WORSENs performance!
//  v0.16B3
//      added message/error log vars&functions for this object
//      added qwLength to struct bit_stream_struc
//  v0.16B24
//     improved the bs_refill_buffer(), doubled buffer-size from 16K -> 32K
//     translate fread, fopen, etc. -> _read, _open, _lseeki64, etc.
//     hopefully _lseeki64 will support >4GB bitstreams
//     added get_fhandle(), returns m_bs.fhandle (handle to FILE)
//  v0.16B23
//     inlined some of the critical functions, to boost performance further
//     changed stream_to_get and substream_to_get to unsigned char
//       #define STREAM_NOT_SET 0
//  v0.16B22
//     added bs_getbytes() function to accelerate reading of packets (>8 bits)
//     added unsigned char m_pgetbytebuffer[], for bs_getbytes() function
//   v0.16B2
//     initial release of C++ object.  Demuxer is now derived from
//     Brent Beyeler's BBDMUX v1.2 (bbdbits is from BBDMUX v1.1)
//
//     The demuxer is used like this :
//     1) declare class cDemuxer oDemuxer;
//     2) oDemuxer.set_stream_id( stream_to_get, substream_to_get );
//     3) oDemuxer.set_filename( Bitstream_file_name );
//     4) oDemuxer.go_demux( &buffer[0], &dwBytesReturned );
//
//     go_demux() returns DMXERR_OK if successful, otherwise a
//        an error code
//
//     The demuxer also records the LBA of the last-known SYSTEM HEADER

/***********************************************************************
*
*  Global Include Files
*
***********************************************************************/


#ifndef _BBDMUX12_H
#define _BBDMUX12_H 1

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
//#include <mem.h>
#include <memory.h>
#include <windows.h>

/* "bit_stream.h" Definitions */

#define DEMUX_MINIMUM         4    // Minimum size of the buffer in bytes
#define DEMUX_MAX_LENGTH      32   // Maximum length of word written or
                                   //     read from bit stream
#define DEMUX_READ_MODE       0
#define DEMUX_WRITE_MODE      1
#define DEMUX_ALIGNING        8
#define DEMUX_BINARY          0
#define DEMUX_ASCII           1
#define DEMUX_BS_FORMAT       BINARY // BINARY or ASCII = 2x bytes
#define DEMUX_BUFFER_SIZE     (65536) + 4	
	// the +4 to compensates for bs_refill's odd refill mechanism
#define DEMUX_LBA_SIZE		  2048 // DVD-disc LBA block size

#define BSERR_OK 0				// no error
#define BSERR_FILE_OPEN	-1		// a bitstream is already open
#define BSERR_FILE_NOT_FOUND -2 // requested bitstream file not found
#define BSERR_FAIL -3			// generic error, no reason stated

#define STREAM_NOT_SET		0 // init vals for stream_to_get,substream_to_get

/* cDemuxer::go_demux() return codes */
#define DMXERR_OK	0
#define DMXERR_ENDOFSTREAM -1	// end of input bitstream
#define DMXERR_NOTMPEG2 -2
#define DMXERR_FAIL	-3
#define DMXERR_OPEN -4		// a file is already open
#define VIDEO_STREAM_ID 	0xE0	// default MPEG-2 streamID for VOB file

#define _NOFILE -1	// error code for _open(), _read(), _close()

#define BSOPEN( X ) ( (X)->fhandle != _NOFILE ) // is bitstream open?


// v0.16B3 the following four defines are for the housekeeping functions/vars
#define TEXTMSG_LENGTH 1024
	// maximum text length of a single log-entry (message or error)
#define LOG_LENGTH 32767
	// maximum text length of messagelog and errorlog
#define LOG_NEW_ENTRY 0x0000001L // flag
#define LOG_RESET 0x0000002L	// flag

/* "bit_stream.h" Type Definitions */
typedef struct bit_stream_struc
{
  __int64 qwLength;  // total length (in #bytes), QWORD (not DWORD!)
  DWORD dwLBAPos;	// LBA position within bitstream, set by bs_update_lba()
  DWORD dwBytePos;	// Byte-position within current LBA, any function which
	// advances the bitstream-buffer (bs->buf), must manually update this
	// counter
  DWORD dwLBAOffset;// Byte-offset (from pos 0) to start of new LBA,
					// set by bs_update_lba()
  char filename[ 1024 ];	// filename 
//  FILE        *pt;			// pointer to bit stream device
  int		  fhandle;		// file handle
  unsigned char *buf;		// bit stream buffer
  int         buf_size;		// size of buffer (in number of bytes)
  double      totbit;		// bit counter of bit stream
  int         buf_byte_idx;	// pointer to top byte in buffer
  int         buf_bit_idx;	// pointer to top bit of top byte in buffer
  int         mode;			// bit stream open in read or write mode
  int         eob;			// end of buffer index
  int         eobs;			// end of bit stream flag
  char        format;

    /* format of file in rd mode (BINARY/ASCII) */
} Bit_stream_struc;

/* demuxer's internal state */
typedef struct demuxestate_struct
{
  unsigned long PES_packet_length, PES_header_data_length;
  unsigned long PTS_DTS_flags, ESCR_flag, ES_rate_flag, DSM_trick_mode_flag;
  unsigned long additional_copy_info_flag, PES_CRC_flag, PES_extension_flag; 
  unsigned long PES_private_data_flag, pack_header_field_flag;
  unsigned long program_packet_sequence_counter_flag, PSTD_buffer_flag;
  unsigned long PES_extension_flag_2, PES_extension_field_length;
  unsigned long bytes_used, encrypted;

  unsigned long stream_packets[256], stream_bytes[256], pack_packets;
  unsigned long system_packets, ac3_bytes;
  unsigned long substream_packets[256], substream_bytes[256];

  // status variables, the cDemuxer sets these LBA vars as the inputfile is
  // dmultiplexed, but otherwise the demuxer does not use them.  They're
  // here for print_summary()
  DWORD dwCurrHeaderLBA;	// LBA# of most recent System-header
  DWORD dwLastHeaderLBA;	// LBA# of previous System-header
  
  bool streams_found[256];		// flags
  bool stream_encrypted[256];	// flags
  bool substream_encrypted[256];// flags
  bool substreams_found[256];	// flags
  bool vob_flag, mpeg2_flag;	// flags
  bool firsttime;	// flags
  
} demuxstate_t;

/***********************************************************************
*
*  Global Function Prototype Declarations
*
***********************************************************************/

class cDemuxer
{
public :
	void print_summary( void );
	int go_demux( unsigned char *buffer, unsigned long *pcbBuffer );
	DWORD reset_demux( void ); // reset the demux's vars
	int set_target_id( unsigned char stream_id, unsigned char substream_id );
	void get_target_id( unsigned char *stream_id, unsigned char *substream_id );
	int set_filename( char *input_filename );
	int fseek_lba( DWORD dwLBA );  // public interface
	DWORD get_system_llba( void );	// public interface
		// returns LBA address of last good System-header
	DWORD lba_seek_next_sysh( DWORD dwStart ); // seek until find LBA
	int get_fhandle( void );	// return file handle of open bitstream

	// constructor and destructor
	cDemuxer( void );
	~cDemuxer();

protected :
	// private functions, bitstream functions
	int            bs_memcheck(char*, int, int);
	void           *bs_mem_alloc(unsigned long, char*);
	void           bs_mem_free(void**);
	DWORD          bs_refill_buffer(Bit_stream_struc*);
	void           bs_empty_buffer(Bit_stream_struc*, int);
	//void           bs_open_bit_stream_w(Bit_stream_struc*, const char*, int);
	//void           bs_open_bit_stream_r(Bit_stream_struc*, const char*, int);
	long           bs_open_bit_stream_r(Bit_stream_struc*, int);
	long           bs_close_bit_stream_r(Bit_stream_struc*);
	void           bs_close_bit_stream_w(Bit_stream_struc*);
	void           bs_alloc_buffer(Bit_stream_struc*, int);
	void           bs_desalloc_buffer(Bit_stream_struc*);
	void           bs_back_track_buffer(Bit_stream_struc*, int);
	unsigned int bs_get1bit(Bit_stream_struc *bs);
	void           bs_put1bit(Bit_stream_struc*, int);
	unsigned long  bs_look_ahead(Bit_stream_struc*, int);
	__inline unsigned long  bs_getbits(Bit_stream_struc*, int);
	void           bs_putbits(Bit_stream_struc*, unsigned int, int);
	void           bs_byte_ali_putbits(Bit_stream_struc*, unsigned int, int);
	unsigned long  bs_byte_ali_getbits(Bit_stream_struc*, int);
	double         bs_sstell(Bit_stream_struc *bs);
	int            bs_end_bs(Bit_stream_struc*);
	int            bs_seek_sync(Bit_stream_struc*, unsigned int, int);
	__inline unsigned long  bs_mod_double(double val, double modval);
	__inline unsigned long  bs_update_lba( Bit_stream_struc *pbs );
	long           bs_fseek_lba( Bit_stream_struc *pbs, DWORD dwLBA );
	void           bs_reset_buffer( Bit_stream_struc *pbs);
	DWORD          bs_get_lba( Bit_stream_struc *pbs, DWORD *lpdword );
	DWORD          bs_getbytes( Bit_stream_struc *pbs, unsigned char pBuffer[],
					DWORD N, DWORD *pcbRead);

	// these functions "process_xxx()" were not part of BBDMUX...I added them
	// to help modularize go_demux().  In the future, if I were to ever add
	// audio-support, I could easily modify process_xxx() to dispatch
	// data from multiple streams.
	// The original BBDMUX does not have this characteristic.  (In other
	// words, extracting multiple streams would require multiple instances
	// of BBDMUX reading the same file, or major modification to the main()
	// routine.)
	__inline unsigned long  process_stream_payload( unsigned char stream_id, 
		unsigned long cbData );
	__inline unsigned long  process_substream_payload( unsigned char stream_id,
		unsigned char substream_id,	unsigned long cbData );
	unsigned long  process_substream( unsigned char stream_id, 
		unsigned char substream_id );

	// s_fwrite() is a wrapper function for exporting the extracted data.
	// The process_xxx() functions use s_fwrite() to push data into cDemuxer's
	// output buffer area.
	size_t s_fwrite( unsigned char *buffer, size_t size, size_t count, FILE *stream );

	// housekeeping functions, text log
	void TextLogErr( DWORD errorID, char text[] );
	void TextLogMsg( DWORD dwFlags, char text[] );

	// private variables

#if BS_FORMAT != BINARY
	static char *he = "0123456789ABCDEF";
#endif
	static unsigned int m_val;
	Bit_stream_struc m_bs; // bitstream state for bbdbits
	demuxstate_t s; // cDemuxer state, ::go_demux()'s local vars
	
	unsigned char stream_to_get;	// target streamID
	unsigned char substream_to_get;	// target substreamID
	FILE *m_outfp;	// unused output file-pointer

	// these two vars are used by s_fwrite(), the fwrite() emulator.
	// fwrite is called by the demuxer (go_demux() ) and its subfunctions
	unsigned long m_cbFwriteBuf;	// counter, #bytes stored in m_cbFwriteBuf
	unsigned char *m_pFwriteBuf;		// point to FwriteBuffer

	unsigned char m_pBuffer[ DEMUX_BUFFER_SIZE ]; // for bs_getbytes()

	//
	// housekeeping variables for logging purposes
	//

	char *m_szErrorLog;	// text log for error messages
	char *m_szMessageLog;	// text log for miscellaneous messages
	char *m_szText;		// scratch space for functions, temp only!
	char *m_szText2;	// more scratch space for functions, temp only!
	char *m_szLastLogErr; // last error message, set by TextLogErr(),
		// reset by TextLogMsg( LOG_NEW_ENTRY, ... )

	DWORD m_logCounter; // message ID counter
	DWORD m_lastError; // last error code, set by TextLogErr(),
};

#endif