#ifndef _DEMUXBUF
#define _DEMUXBUF 1
#include <windows.h>

/*
 * Simple queue to integrated MPEG2V12 with BBDMUX 
 ******************************************************
 *
 * This DEMUXqueue holds data of type (unsigned char[]).  It sits
 * between the MPEG2V12 getbits() package and BBDMUX's stream
 * output ( fwrite() ) system.
 *
 * QUEUE operation ( class demux_buffer )
 * --------------------------------------------
 *
 * This is a very rudimentary (FIFO) queue for bytedata.  
 * 
 * The queue's input and output are accessed through
 * push_data() and pull_data() 
 *
 * push_data() will report QUEUE_FULL if the queue's storage overfills
 * pull_data() will report FILL_NODATA if the queue underruns
 *
 * otherwise, either function will report the #bytes actually stored/retrieved
 *
 * There is one "twist" to this queue.  In addition to push/pull
 * operation, the queue can be set to "DMX_MODE_AUTOFILL" mode.  The user 
 * supplies a "fill-handler" (a function-call), and the queue automatically 
 * calls the fill-handler (to refill the queue) as necessary.
 *
 *
 * BBDMUX and QUEUEing programming issues 
 * --------------------------------------------
 *
 * A standard queue works in complimentary "push/pull" fashion : the
 * user puts data into the queue for storage, then extracts data as necessary.
 * BBDMUX is loop-driven.  A while() loop executes until the input-file 
 * reaches EOF.  Clearly, if BBDMUX's output were connected to the 
 * demux_queue's input, without any sort of flow-control the queue would
 * saturate.  
 *
 * To get around this, I changed the queue's from "complimentary push/pull"
 * to "demand (pull) driven."  It is still a queue, except the assumption
 * is that MPEG2V12 will hammer the queue's pull_data() function.
 *
 * When the queue empties (or runs low), pull_data() automatically calls
 * a fill_queue() to refill the queue.  The fill_queue() function is
 * responsible for refilling the queue.  No part of the application (MPEG2V12)
 * directly calls push_data().  Thus, the queue is essentially "demand driven."
 *
 * To support this queuing mechanism, I modified BBDMUX so that it executes
 * the while() loop once, then exits.  (Actually, the while() loop runs until
 * it returns >0 bytes of data.  Under normal conditions the loop could execute
 * anywhere from 0 to 4-5 times before returning to caller.)
 *
 *
 * As MPEG2V12's getbits() pulls data from DEMUXqueue, the queue
 * automatically refills (when the amount requested exceeds the amount
 * stored in the queue.)  The "refill" mechanism determines the
 * queue's mode of operation.
 *
 * Revision History
 * ----------------------------
 * v0.16B2 updated the demux-buffer's fill-handler definition, (added
 *         ULONG *)
 * v0.13 08/09/99 first release 
 *
 *
 * liaor@stanford.edu (liaor@iname.com)
 *
 */

#define DMX_BUFFER_SIZE	65536
#define DMX_FILLTHRESHOLD 32768
#define FILL_NODATA -1
#define FILL_QUEUEFULL -2			// fatal error
#define FILL_QUEUEOVERTHRESHOLD -3	// warning
#define FILL_OK 0

#define DMX_MODE_AUTOFILL 1
#define DMX_MODE_NOAUTOFILL 2

typedef long (*handlerFunc)( unsigned char[], unsigned long * );
	// user handler function
	// "handlerFunc" is now of type (*)(...)

class demux_buffer
{
public: 
	// constructor & destructor
	demux_buffer( void );
	~demux_buffer();

	// public member functions (public interface)
	long get_qsize( void )
		{ return cbstored; };

	long get_capacity( void )
	{ return DMX_BUFFER_SIZE; };

	long push_data( LPBYTE indata, long cbSize );
	long pull_data( LPBYTE outdata, long cbrequest );
	void reset( void );

	int set_autofill_handler( handlerFunc );
	int set_autofill_mode( int new_mode );
	

	LPBYTE public_buffer;	// public buffer, returned as parameter

private :	
	LPBYTE buffer_memory;	// internal storage
	
	long in_pos;	// index to *TOP* of queue (incoming data written
					// to top of queue 
	long out_pos;	// index to *BOTTOM* of queue (data extracted from bottom)
	long cbstored;	// byte-count of total #bytes-stored in this queue

	long fill_queue( long cbrequest ); // "pull" data from the source, calls push_data()
				// if (mode == AUTOFILL), fill_queue() is called after an
				// underflow condition

	long (*autofill_handler)( unsigned char *buffer, unsigned long *cbBuffer);

	int mode;	// stores DMX_MODE, specifies queue's policy when a pull-
				// operation empties the queue.  The queue can be setup to
				// call an external function when pull_data() causes a
				// queue_underflow condition.  
};
#endif
