/* store.cpp, picture output routines                                         */

/* Copyright (C) 1996, MPEG Software Simulation Group. All Rights Reserved. */

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  The MPEG Software Simulation Group disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 * The MPEG Software Simulation Group does not represent or warrant that the
 * programs furnished hereunder are free of infringement of any third-party
 * patents.
 *
 * Commercial implementations of MPEG-1 and MPEG-2 video, including shareware,
 * are subject to royalty fees to patent holders.  Many of these patents are
 * general enough such that they are unavoidable regardless of implementation
 * design.
 *
 *
 * MPEG2AVI
 * --------
 * v0.16B1 store.c --> store.cpp
 *       upgraded to C++
 * v0.15B51 the UserOptFRangeEnd triggers -> g_cmd_buttonStop 
 * v0.15B5 restructured post-processing operations, fixed malloc() bug in
 *       store_avi_gdi()
 *
 *       | MPEG2V12 | -> | VRESIZE.C | -> | YUV2YUV | -> | CROP&CSC | -> OUTPUT
 *       MPEG decoder    -3X, -3Y resize   pixFormat     crop viewport
 *                                         conversion
 *
 * v0.15B4 direct-YUV AVI mode.  The decoded-MPEG frames are transferred
 *       to storeavi.c in the native YCbCr420 color-format
 *       half-resolution mode now works for all CHROMA formats (420/422/444)
 * v0.15 added frame-counter, g_FrameCount.  This counts the #frames received
 *       from the MPEG-decoder, and is not 100% reliable.  It's a useful
 *       statistic for the GDI and AVI functions, though.
 *       
 *       added process_filter_cmd( void ) to proces g_cmd_filter
 *
 * v0.14
 *    Clear all memory-spaces, 
 *    Created a new function store_avi_gdi() ... 
 *    half-resolution output mode (currently only supports CHROMA420 foramt.)
 *    The 1/2-res mode only supports YUV420 planar, MPEG's most popular
 *    color format.
 *
 *     Cropping & 1/2-resolution implementation
 *     ----------------------------------------
 *
 *     1)  decoder-output <src[0],src[1],src[2]>
 *     2)  Half-resolution?  if yes, downsample
 *     3)  build YUV888 24bpp packed bitmap from <Resized image, YUV444 planar)
 *         (crop & convert image simultaneously)
 *     4)  call output functions (pass YUV888 24bpp packed image)
 *     5)  (display_dib or store_avi_one), convert color space with yuv2rgb24() 
 *         yuv2rgb24()...
 *     6)  done
 *
 *     only store_avi_gdi() implement cropping/downsampling...the other
 *     functions ignore these operations.
 *
 *    When 1/2-res mode is in use, store_avi_gdi() *directly* calls 
 *    ivtc_half_8bpp() [ telecine.c ], then sends the processed frame to
 *    the appropriate output-routine ( store_avi() or display_dib() )
 *
 *    By handling the telecine-conversion in native YUV420 planar, the
 *    upscaling/downscaling (YUV420->422->444) filters are bypassed entirely,
 *    increasing both performance and precision.
 *
 *    Overall data-flow from MPEG-decoder's output to file
 *
 *    1)  store_one()  -> src[0],src[1],src[2] -> store_avi()
 *    2)  src[0]... -> src[0] (y_half), src[1] ->u444, src[2] -> v444 )
 *    3)     fb_yuv -> ivtc_half_8bpp() -> (display_dib or store_avi_one) -> yuv2rgb24()
 *
 * v0.11
 * Most modifications are to store_ppm_tga() ... in addition to its previous
 * PPM/TGA output, this function now handles GDI and AVI.
 *
 * store_ppm_tga() redirects GDI output to display_dib() [windisp.c]
 *					 and AVI to store_avi_one() [storeavi.c]
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <io.h>

#include "config.h"
#include "global.h"
#include "storeavi.h"
#include "yuv2yuv.h"
#include <windows.h> /* used only to get type declaration BOOL */

/* private prototypes */
static void store_one _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height));
static void store_yuv _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height));
static void store_sif _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height));
static void store_ppm_tga _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height, int tgaflag));
static void store_avi_gdi _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height, int tgaflag));
static void store_yuv1 _ANSI_ARGS_((char *name, unsigned char *src,
  int offset, int incr, int width, int height));
static void putbyte _ANSI_ARGS_((int c));
static void putword _ANSI_ARGS_((int w));
BOOL process_filter_cmd( void ); /* process command-variable g_cmd_filter */
//static long g_FrameCount;
static RECT srcRect; /* resolution of MPEG-video */

#define OBFRSIZE 4096
static unsigned char obfr[OBFRSIZE];
static unsigned char *optr;
static int outfile;

static int
rstrcmpi( const char a[], const char b[] )
{
	long counter = 0;
	long result = 0;
	long len_a = 0, len_b = 0;
	long minab = 0;
	
	if ( a != NULL )
		len_a = strlen( a );

	if ( b != NULL )
		len_b = strlen( b );

	if ( len_a == 0 && len_b == 0 )
		return 0;	/* equal ?!? */

	if ( len_a == 0 )
		return -1; /* left < right */

	if ( len_b == 0 )
		return 1; /* left > right */

	minab = ( len_a > len_b ) ? len_b : len_a;

	for ( counter = 0; counter < minab && result == 0; ++counter )
	{
		if ( toupper( a[ len_a - counter ] ) !=
			toupper( b[ len_b - counter ] ) )
			result = toupper( a[ len_a - counter ] ) >
				toupper( b[ len_b - counter ] ) ? 1 : -1 ;
	}

	return result;
}

/*
 * store a picture as either one frame or two fields
 */
void Write_Frame(unsigned char *src[],int frame)
{
  char outname[FILENAME_LENGTH];

  if (progressive_sequence || progressive_frame || Frame_Store_Flag)
  {
    /* progressive */
    sprintf(outname,Output_Picture_Filename,frame,'f');
    store_one(outname,src,0,Coded_Picture_Width,vertical_size);
  }
  else
  {
    /* interlaced */
    sprintf(outname,Output_Picture_Filename,frame,'a');
    store_one(outname,src,0,Coded_Picture_Width<<1,vertical_size>>1);

    sprintf(outname,Output_Picture_Filename,frame,'b');
    store_one(outname,src,
      Coded_Picture_Width,Coded_Picture_Width<<1,vertical_size>>1);
  }

  //++g_FrameCount;
}

/*
 * store one frame or one field
 */
static void store_one(char outname[],unsigned char *src[], int offset, 
	int incr, int height)
{
  static int first_iteration = 1;
  if ( first_iteration )
  {
	  first_iteration = 0;
	  //g_FrameCount = 0;  /* initialize frame counter */
	  printf("\nMPEG-%d video resolution : <%d x %d>, coded format = ",
		  (1+base.MPEG2_Flag), Coded_Picture_Width,Coded_Picture_Height );
	  switch( chroma_format )
	  {
	    case CHROMA420 :
		  printf("YUV420");
		  break;
	    case CHROMA422 :
		  printf("YUV422");
		  break;
	    case CHROMA444 :
		  printf("YUV444");
		  break;
	    default:
		  printf("unknown!");
	  } /* endswitch */

	  /* setup rectangle for store.c */
	  srcRect.left = 0; 
	  srcRect.top = 0;
	  srcRect.right = Coded_Picture_Width;
	  srcRect.bottom = Coded_Picture_Height;
	  printf("\n");
  } /* endif ( first_iteration ) */
  
  switch (Output_Type)
  {
  case T_YUV:
    store_yuv(outname,src,offset,incr,height);
    break;
  case T_SIF:
    store_sif(outname,src,offset,incr,height);
    break;
  case T_TGA:
    store_ppm_tga(outname,src,offset,incr,height,1);
    break;
  case T_PPM:
    store_ppm_tga(outname,src,offset,incr,height,0);
    break;
#ifdef DISPLAY
  case T_X11:
    dither(src);
    break;
#endif
  case T_GDI:
	store_avi_gdi(outname,src,offset,incr,height,T_GDI);
	break;
  case T_AVI:
	store_avi_gdi(outname,src,offset,incr,height,T_AVI);
	break;
  case T_AVI_YV12:  // quick hack for 0.15B4
	store_avi_gdi(outname,src,offset,incr,height,T_AVI_YV12);
	break;
  case T_AVI_YUY2:  // quick hack for 0.15B5
	store_avi_gdi(outname,src,offset,incr,height,T_AVI_YUY2);
	break;
  default:
	  fprintf(stderr, "\nstore_one() : Unknown output mode -o%d", Output_Type );
	  exit(ERROR);
    break;
  }
  
  ++g_cbMPEGPicture; /* update decoded picture counter */

}

/* separate headerless files for y, u and v */
static void store_yuv( char outname[], unsigned char *src[], int offset, 
	int incr, int height )
{
  int hsize;
  char tmpname[FILENAME_LENGTH];

  hsize = horizontal_size;

  sprintf(tmpname,"%s.Y",outname);
  store_yuv1(tmpname,src[0],offset,incr,hsize,height);

  if (chroma_format!=CHROMA444)
  {
    offset>>=1; incr>>=1; hsize>>=1;
  }

  if (chroma_format==CHROMA420)
  {
    height>>=1;
  }

  sprintf(tmpname,"%s.U",outname);
  store_yuv1(tmpname,src[1],offset,incr,hsize,height);

  sprintf(tmpname,"%s.V",outname);
  store_yuv1(tmpname,src[2],offset,incr,hsize,height);
}

/* auxiliary routine */
static void store_yuv1( char name[], unsigned char *src, int offset,
	int incr, int width, int height )
{
  int i, j;
  unsigned char *p;

  if (!Quiet_Flag)
    fprintf(stderr,"saving %s\n",name);

  if ((outfile = _open(name,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",name);
    Error(Error_Text);
  }

  optr=obfr;

  for (i=0; i<height; i++)
  {
    p = src + offset + incr*i;
    for (j=0; j<width; j++)
      putbyte(*p++);
  }

  if (optr!=obfr)
    _write(outfile,obfr,optr-obfr);

  _close(outfile);
}

/*
 * store as headerless file in U,Y,V,Y format
 */
static void store_sif (char outname[], unsigned char *src[], int offset,
	int incr, int height )
{
  int i,j;
  unsigned char *py, *pu, *pv;
  static unsigned char *u422, *v422;

  if (chroma_format==CHROMA444)
    Error("4:4:4 not supported for SIF format");

  if (chroma_format==CHROMA422)
  {
    u422 = src[1];
    v422 = src[2];
  }
  else
  {
    if (!u422)
    {
      if (!(u422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                           *Coded_Picture_Height)))
        Error("malloc failed");
      if (!(v422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                           *Coded_Picture_Height)))
        Error("malloc failed");
    }
  
    conv420to422(src[1],src[2],u422,v422, Coded_Picture_Width,
		Coded_Picture_Height );
  }

  strcat(outname,".SIF");

  if (!Quiet_Flag)
    fprintf(stderr,"saving %s\n",outname);

  if ((outfile = _open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",outname);
    Error(Error_Text);
  }

  optr = obfr;

  for (i=0; i<height; i++)
  {
    py = src[0] + offset + incr*i;
    pu = u422 + (offset>>1) + (incr>>1)*i;
    pv = v422 + (offset>>1) + (incr>>1)*i;

    for (j=0; j<horizontal_size; j+=2)
    {
      putbyte(*pu++);
      putbyte(*py++);
      putbyte(*pv++);
      putbyte(*py++);
    }
  }

  if (optr!=obfr)
    _write(outfile,obfr,optr-obfr);

  _close(outfile);
}

/*
 * store as PPM (PBMPLUS) or uncompressed Truevision TGA ('Targa') file
 */
static void store_ppm_tga( char outname[], unsigned char *src[], int offset,
	int incr, int height, int tgaflag )
{
  int i, j;
  int y, u, v, r, g, b;
  int crv, cbu, cgu, cgv;
  unsigned char *py, *pu, *pv;
  static unsigned char tga24[14] = {0,0,2,0,0,0,0, 0,0,0,0,0,24,32};
  char header[FILENAME_LENGTH];
  unsigned char *pixelxy_addr = NULL;
  static unsigned char *u422, *v422, *u444, *v444;
  
  
  if (chroma_format==CHROMA444)
  {
    u444 = src[1];
    v444 = src[2];
  }
  else
  {
    if (!u444)
    {
      if (chroma_format==CHROMA420)
      {
        if (!(u422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                             *Coded_Picture_Height)))
          Error("malloc failed");
        if (!(v422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                             *Coded_Picture_Height)))
          Error("malloc failed");
      }

      if (!(u444 = (unsigned char *)malloc(Coded_Picture_Width
                                           *Coded_Picture_Height)))
        Error("malloc failed");

      if (!(v444 = (unsigned char *)malloc(Coded_Picture_Width
                                           *Coded_Picture_Height)))
        Error("malloc failed");
    }

    if (chroma_format==CHROMA420)
    {
      conv420to422(src[1],src[2],u422,v422, Coded_Picture_Width, 
		  Coded_Picture_Height );
      conv422to444(u422,v422,u444,v444,Coded_Picture_Width, 
		  Coded_Picture_Height );
    }
    else
    {
      conv422to444(src[1],src[2],u444,v444,Coded_Picture_Width, 
		  Coded_Picture_Height );
    }
  }

  /*strcat(outname,tgaflag ? ".tga" : ".ppm");*/
  switch ( tgaflag )
  {
	case 0 : if ( rstrcmpi( outname, ".ppm" ) != 0 )
				 strcat( outname, ".ppm" );
		break;
	case 1 : if ( rstrcmpi( outname, ".tga" ) != 0 )
				 strcat( outname, ".tga" );
		break;
	default :
		;
  }
  

  if ( tgaflag )
  {
  if (!Quiet_Flag)
    fprintf(stderr,"saving %s\n",outname);

  if ((outfile = _open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",outname);
    Error(Error_Text);
  }
  }

  optr = obfr;

  if (tgaflag)
  {
    /* TGA header */
    for (i=0; i<12; i++)
      putbyte(tga24[i]);

    putword(horizontal_size); putword(height);
    putbyte(tga24[12]); putbyte(tga24[13]);
  }
  else
  {
    /* PPM header */
    sprintf(header,"P6\n%d %d\n255\n",horizontal_size,height);

    for (i=0; header[i]!=0; i++)
      putbyte(header[i]);
  }

  /* matrix coefficients */
  crv = Inverse_Table_6_9[matrix_coefficients][0];
  cbu = Inverse_Table_6_9[matrix_coefficients][1];
  cgu = Inverse_Table_6_9[matrix_coefficients][2];
  cgv = Inverse_Table_6_9[matrix_coefficients][3];
  
  for (i=0; i<height; i++)
  {
    py = src[0] + offset + incr*i;
    pu = u444 + offset + incr*i;
    pv = v444 + offset + incr*i;

    for (j=0; j<horizontal_size; j++)
    {
      u = *pu++ - 128;
      v = *pv++ - 128;
      y = 76309 * (*py++ - 16); /* (255/219)*65536 */
      r = Clip[(y + crv*v + 32768)>>16];
      g = Clip[(y - cgu*u - cgv*v + 32768)>>16];
      b = Clip[(y + cbu*u + 32786)>>16];
	  
	  if (tgaflag)
	  {
	        putbyte(b); putbyte(g); putbyte(r);
	  }
	  else
	  {
	        putbyte(r); putbyte(g); putbyte(b);
	  }
    }
  }

  if (optr!=obfr)
  {
    _write(outfile,obfr,optr-obfr); /* write image to file */
    _close(outfile);
  }
}

/*
 * store as AVI (Windows) or display to Windows GDI
 */
static void store_avi_gdi( char outname[], unsigned char *src[], int offset,
	int incr, int height, int tgaflag )
{
  int i, j;
  /*int y, u, v, r, g, b;*/
  DWORD y_src_addr, uv_src_addr, dest_addr;
  RECT dstRect;
  DWORD dwPixFormat = 0;	/* needed for avi_fileopen() */
  
  static unsigned char *u422 = NULL, *v422= NULL, *u444=NULL, *v444=NULL,
	  *y_full=NULL, *u_extra = NULL, *v_extra = NULL, *y_extra = NULL;
  
  static int first_iteration = 1;
  static unsigned char *fb_yuv = NULL, *y_ptr = NULL , *u_ptr = NULL,
	  *v_ptr = NULL;
  static long m_OutputHeight, m_OutputWidth; 
	/* dimensions of DISPLAYED (final) frames, after cropping */

  static RECT CropRect; /* cropping rectangle, passed to -o6 and -o7 */
  /* the cropping rectangle is with respect to Resized_Picture_xxx
     yuv2rgb24 ultimately handles the cropping. */
  tImageOpBlock convBlock; /* conversion information for yuv2yuv() */
  
  if ( first_iteration )
  {
	first_iteration = 0;	/* disable this flag */

  	/* allocate memory and clear it */
	fb_yuv = (unsigned char *)malloc(   Coded_Picture_Width * 
		Coded_Picture_Height * 3); 
	y_full = (unsigned char *)malloc(   Coded_Picture_Width * 
		Coded_Picture_Height );
	/* y, u and v extra are for the resizer */
	y_extra = (unsigned char *)malloc(   Coded_Picture_Width * 
		Coded_Picture_Height );
	u_extra = (unsigned char *)malloc(   Coded_Picture_Width * 
		Coded_Picture_Height );
	v_extra = (unsigned char *)malloc(   Coded_Picture_Width * 
		Coded_Picture_Height );
	u422 = (unsigned char *)malloc( Coded_Picture_Width*
		Coded_Picture_Height);
	v422 = (unsigned char *)malloc( Coded_Picture_Width*
		Coded_Picture_Height);
	u444 = (unsigned char *)malloc( Coded_Picture_Width*
		Coded_Picture_Height);
	v444 = (unsigned char *)malloc( Coded_Picture_Width*
		Coded_Picture_Height);

	if ( !( fb_yuv && y_full && y_extra && u_extra && v_extra &&
		u422 && v422 && u444 && v444 ) )
	{
		fprintf( stderr, 
			"\nstore_avi_gdi() : could not allocate temp framebuffers!");
		exit( ERROR );
	}
	else
	{
		/* clear the framebuffers! */
		memset((LPVOID)fb_yuv,0,Coded_Picture_Width*Coded_Picture_Height*3);
		memset((LPVOID)y_full,0,Coded_Picture_Width*Coded_Picture_Height);
		memset((LPVOID)y_extra,0,Coded_Picture_Width*Coded_Picture_Height);
		memset((LPVOID)u_extra,128,Coded_Picture_Width*Coded_Picture_Height);
		memset((LPVOID)v_extra,128,Coded_Picture_Width*Coded_Picture_Height);
		memset((LPVOID)u422,128,Coded_Picture_Width*Coded_Picture_Height);
		memset((LPVOID)v422,128,Coded_Picture_Width*Coded_Picture_Height);
		memset((LPVOID)u444,128,Coded_Picture_Width*Coded_Picture_Height);
	  	memset((LPVOID)v444,128,Coded_Picture_Width*Coded_Picture_Height);
	}  /* endif */

	if ( tgaflag != T_GDI )
	{
		if ( rstrcmpi( outname, ".avi" ) != 0 )
			strcat( outname, ".avi" );
	}

	/* check chroma format if ( g_vHalfResFlag == TRUE )
	   currently we only support CHROMA420 for half-resolution mode */
	//if ( g_vHalfResFlag && chroma_format != CHROMA420 )
	//{
	//	fprintf( stderr, "\nstore.c - half-resolution requires color-format of input bitstream to be CHROMA420." );
	//	  Error("Color format incompatible");
	  //}

	/* load valid co-ordinates into Resized_Picture_Width/Height */

	Resized_Picture_Width= g_vHalfResFlag ? (Coded_Picture_Width >> 1 ):
		Coded_Picture_Width;
	Resized_Picture_Height = g_vHalfResFlag ? (Coded_Picture_Height >> 1):
		Coded_Picture_Height;

	/* check if g_AVIcrop already has valid co-ordinates, if not set
	   them to current image-height/width...i.e. effectively disables cropping */
    /* m_OutputWidth, m_OutputHeight replace g_AVIcrop... */
	m_OutputWidth = Resized_Picture_Width;
	m_OutputHeight = Resized_Picture_Height;

	if ( CHECKOPT( UserOptCropX ) )
		m_OutputWidth = GETOPT( UserOptCropX );

	if ( CHECKOPT( UserOptCropY ) )
		m_OutputHeight = GETOPT( UserOptCropY );

	/* validate g_AVIcrop & Resized_Picture ... are parameters legal? */
	/* Make sure cropped sized is smaller than resized-image */
	if ( m_OutputWidth > Resized_Picture_Width || 
		m_OutputHeight > Resized_Picture_Height )
	{
		fprintf( stderr, "\nstore_avi_gdi() : Viewport dimensions (\"-1\" %ld X %ld)",
			m_OutputWidth , m_OutputHeight );
		fprintf( stderr, " exceed\n\tresized-output's dimensions (%ld X %ld) !",
			Resized_Picture_Width, Resized_Picture_Height );
		exit (ERROR );
	}
				

	CropRect.top = ( Resized_Picture_Height - m_OutputHeight  ) >> 1;
	CropRect.bottom = Resized_Picture_Height - 
		(( Resized_Picture_Height - m_OutputHeight ) >> 1);
	CropRect.left = ( Resized_Picture_Width - m_OutputWidth ) >> 1;
	CropRect.right = Resized_Picture_Width -
		(( Resized_Picture_Width - m_OutputWidth ) >> 1 );

	if ( tgaflag != T_GDI )
	{
		switch ( Output_Type )
		{
		case T_AVI_YV12 :
//			fUseYUV = TRUE;
			dwPixFormat = mmioFOURCC( 'Y','V','1','2');
			break;
		case T_AVI_YUY2 :
//			fUseYUV = TRUE;
			dwPixFormat = mmioFOURCC( 'Y','U','Y','2');
			break;
		default :
			dwPixFormat = BI_RGB;
//			fUseYUV = FALSE;
		} /* endswitch ( Output_Type ) */

		avi_fileopen( outname, &CropRect, dwPixFormat );
	}


  } /* endif ---- if ( first_iteration ) */

  process_filter_cmd( );  /* process commands to filter routines */
  process_resize_cmd( );  /* process commands to resize routines */
	
  /* check UserOptFRangeStart & UserOptFRangeEnd */

  /* gethdr.c skips B-frames, but we still need to skip I/P frame processing
     we can't simply skip I/P frames entirely, because we might start
	 displaying on a B-frame which depends on previous I/P frames! */
  /* the inequalities imply that <a> ... <b> are INCLUDED in the range! */
  if ( CHECKOPT( UserOptFRangeStart ) )
  {
	  if ( GETOPT( UserOptFRangeStart ) >= g_cbMPEGPicture)
	  {
		  return;
	  }
  }

  if ( CHECKOPT( UserOptFRangeEnd ) )
  {
	 if ( g_cbMPEGPicture > GETOPT( UserOptFRangeEnd ) )
	 {
		 printf("\r< FRangeEnd > = MPEG-FrameCount (%ld)>", g_cbMPEGPicture );
		 g_cmd_buttonStop = TRUE;  /* stop after current frame */
	 }
  }

  y_ptr = src[0];
  u_ptr = src[1];
  v_ptr = src[2];

//  if ( !g_vHalfResFlag ) /* no 1/2 downsampling, FULL resolution */
//  {
	//y_ptr = ivtc_full_8bpp( src[ 0 ], g_FrameCount );
  y_ptr = ivtc_full_8bpp( src[ 0 ], g_cbMPEGPicture );
	
	/* Resize the output image */
	if ( y_ptr != NULL )  /* did ivtc delete current frame? */
	{ /* current frame wasn't deleted, resize the output! */
		if ( CHECKOPT( UserOptResizeX ) || CHECKOPT( UserOptResizeY ) )
		{
			dstRect.top = 0;
			dstRect.left = 0;
			dstRect.right = CHECKOPT( UserOptResizeX ) ? UserOptResizeX.data :
				srcRect.right;
			dstRect.bottom = CHECKOPT( UserOptResizeY ) ? UserOptResizeY.data :
				srcRect.bottom;
			shrink_frame( y_extra, u_extra, v_extra, y_ptr, u_ptr, v_ptr,
				&srcRect, &dstRect, chroma_format );

			  /* don't forget that RGB24 requires CHROMA444 to function */

			y_ptr = y_extra;
			u_ptr = u_extra;
			v_ptr = v_extra;
		}
	} /* endif ( y_ptr != NULL ) */

	/* at this point, the framebuffer y_ptr, u_ptr, v_ptr holds a resized
	   bitmap, whose format == chroma_format (specified by MPEG-bistream) */
  
  if ( y_ptr != NULL )  /* did ivtc delete current frame? */
  { /* current frame wasn't deleted, convert ouput format */
		convBlock.srcRect = srcRect;
		// convBlock.destRect = not used
		convBlock.dwSrcFormat = chroma_format;
		switch ( tgaflag )
		{
		  case T_AVI_YV12 :
			convBlock.dwDstFormat = CHROMA420;
			break;
		  case T_AVI_YUY2 :
			convBlock.dwDstFormat = CHROMA422;
			break;
		  default : /* RGb24 */
			convBlock.dwDstFormat = CHROMA444;
			break;
		} /* endswitch */
		
		/* specify input buffers */
		convBlock.src1 = y_ptr;
		convBlock.src2 = u_ptr;
		convBlock.src3 = v_ptr;
		/* supply allocated buffer space for operation output*/
		convBlock.dst1 = y_full;
		convBlock.dst2 = u444;
		convBlock.dst3 = v444;
		yuv2yuv( &convBlock );
		y_ptr = convBlock.outdst1;
		u_ptr = convBlock.outdst2;
		v_ptr = convBlock.outdst3;

	switch ( tgaflag )
	{
	  case T_AVI_YV12 : /* YV12 planar 12bpp */
		/* convert & crop YUV4:2:0 planar to YV12 12bpp planar format */
		/* framebuffer[] = YV12 planar image */
		/* do not need to invert image */
		for (j=0; j<m_OutputHeight ; ++j )
		{
			y_src_addr = (j+CropRect.top)*Resized_Picture_Width + CropRect.left;
			dest_addr = j * m_OutputWidth;
			memcpy( fb_yuv + dest_addr, y_ptr + y_src_addr, m_OutputWidth);
		}

		for ( j=0; j< (m_OutputHeight>>1); ++j )
		{
			dest_addr = ( m_OutputHeight * m_OutputWidth) +
				(j * (m_OutputWidth>>1) );
			uv_src_addr = (j+(CropRect.top>>1))*(Resized_Picture_Width>>1) + 
				(CropRect.left>>1);
			memcpy( fb_yuv+dest_addr, v_ptr+uv_src_addr, (m_OutputWidth>>1) );
	
			dest_addr += (m_OutputHeight>>1) * (m_OutputWidth>>1);
			//src_addr = (j+CropRect.top)*(Resized_Picture_Width>>1) + 
			//	(CropRect.left>>1);
			memcpy( fb_yuv+dest_addr, u_ptr+uv_src_addr, (m_OutputWidth>>1) );
		} /* endfor */
		break;

	  case T_AVI_YUY2 : /* YUY2 packed 16bpp */
		/* convert & crop YUV4:2:2 planar to YUY2 packed 16bpp format */
		/* fb_yuv[] = YUY2 planar image */
		/* do not need to invert image */
		for (j=0; j<m_OutputHeight ; ++j )
		{
			for ( i=0; i< (m_OutputWidth>>1); ++i )
			{	/* i = "group index" (one group is a set of four pixels [YUYV] */
				dest_addr = (j<<1) * m_OutputWidth + (i<<2);
				y_src_addr = (j+CropRect.top)*Resized_Picture_Width + 
					CropRect.left + (i<<1);
				uv_src_addr = (j+CropRect.top)*(Resized_Picture_Width>>1) + 
					(CropRect.left>>1) + i;
				*(fb_yuv+dest_addr+ 0) = *(y_ptr + y_src_addr  );	/* 0) Y */
				*(fb_yuv+dest_addr+ 1) = *(u_ptr + uv_src_addr );	/* 1) U */
				*(fb_yuv+dest_addr+ 2) = *(y_ptr + y_src_addr +1 );	/* 2) Y */
				*(fb_yuv+dest_addr+ 3) = *(v_ptr + uv_src_addr );	/* 3) V */
			}
		} /* endfor ( j=0 ... ) */
		break;

	  default: /* RGB24 */
		/* convert & crop YUV4:4:4 planar to YUV 24bpp packed pixel */
		/* note we do *NOT* perform color-space conversion just yet! */
		/* framebuffer[] = YUV 24bpp packed pixel image */
		/* invert image at the same time */

		for (j=0; j<m_OutputHeight ; j++)
		{
			for (i=0; i<m_OutputWidth ; i++)
			{
				y_src_addr = (j+CropRect.top)*Resized_Picture_Width +
					(i+CropRect.left);
				/* the "-1" is necessary for proper alignment */
				dest_addr = ( m_OutputHeight - j -1 ) * m_OutputWidth + i;
				*( fb_yuv + 3*dest_addr + 0 ) = *(y_ptr + y_src_addr );
				*( fb_yuv + 3*dest_addr + 1 ) = *(u_ptr + y_src_addr );
				*( fb_yuv + 3*dest_addr + 2 ) = *(v_ptr + y_src_addr );
			}
		}  /* endfor */
	} /* endswitch ( tgaflag ) */


	if ( tgaflag == T_GDI )
	{
		//display_dib( fb_yuv, &CropRect, g_FrameCount ); /* display frame to GDI */
		display_dib( fb_yuv, &CropRect, g_cbMPEGPicture ); /* display frame to GDI */
	}
	else if ( tgaflag == T_AVI || tgaflag == T_AVI_YV12 || tgaflag==T_AVI_YUY2 )
	{
		//store_avi_one( outname, fb_yuv, &CropRect, g_FrameCount ); 
		store_avi_one( outname, fb_yuv, &CropRect, g_cbMPEGPicture ); 
			/* write one-frame to AVI */
	}
  } /* endif - if ( y_ptr != NULL ) */

  y_ptr = NULL;
  /*++g_FrameCount; moved to store_one() */

}

static void putbyte(int c)
{
  *optr++ = c;

  if (optr == obfr+OBFRSIZE)
  {
    _write(outfile,obfr,OBFRSIZE);
    optr = obfr;
  }
}

static void putword(int w)
{
  putbyte(w); putbyte(w>>8);
}
